#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from osc_lib.command import command
from osc_lib import utils
from tackerclient.i18n import _
from tackerclient.osc import sdk_utils
from tackerclient.osc import utils as tacker_osc_utils

_VNF_LCM_OP_OCC_ID = 'vnf_lcm_op_occ_id'

_MIXED_CASE_FIELDS = ['operationState', 'stateEnteredTime', 'startTime',
                      'vnfInstanceId', 'grantId', 'isAutomaticInvocation',
                      'isCancelPending', 'cancelMode', 'operationParams',
                      'resourceChanges', 'changedInfo',
                      'changedExtConnectivity']

_FORMATTERS = {
    'operationParams': tacker_osc_utils.FormatComplexDataColumn,
    'error': tacker_osc_utils.FormatComplexDataColumn,
    'resourceChanges': tacker_osc_utils.FormatComplexDataColumn,
    'changedInfo': tacker_osc_utils.FormatComplexDataColumn,
    'changedExtConnectivity': tacker_osc_utils.FormatComplexDataColumn,
    '_links': tacker_osc_utils.FormatComplexDataColumn
}

_ATTR_MAP = (
    ('id', 'id', tacker_osc_utils.LIST_BOTH),
    ('operationState', 'operationState', tacker_osc_utils.LIST_BOTH),
    ('vnfInstanceId', 'vnfInstanceId', tacker_osc_utils.LIST_BOTH),
    ('operation', 'operation', tacker_osc_utils.LIST_BOTH)
)


def _get_columns(vnflcm_op_occ_obj, action=None):

    column_map = {
        'id': 'ID',
        'operationState': 'Operation State',
        'stateEnteredTime': 'State Entered Time',
        'startTime': 'Start Time',
        'vnfInstanceId': 'VNF Instance ID',
        'operation': 'Operation',
        'isAutomaticInvocation': 'Is Automatic Invocation',
        'isCancelPending': 'Is Cancel Pending',
        'error': 'Error',
        '_links': 'Links'
    }

    if action == 'show':
        column_map.update(
            {'operationParams': 'Operation Parameters',
             'grantId': 'Grant ID',
             'resourceChanges': 'Resource Changes',
             'changedInfo': 'Changed Info',
             'cancelMode': 'Cancel Mode',
             'changedExtConnectivity': 'Changed External Connectivity'}
        )

    return sdk_utils.get_osc_show_columns_for_sdk_resource(vnflcm_op_occ_obj,
                                                           column_map)


class RollbackVnfLcmOp(command.Command):
    def get_parser(self, prog_name):
        """Add arguments to parser.

        Args:
            prog_name ([type]): program name

        Returns:
            parser([ArgumentParser]):
        """
        parser = super(RollbackVnfLcmOp, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_LCM_OP_OCC_ID,
            metavar="<vnf-lcm-op-occ-id>",
            help=_('VNF lifecycle management operation occurrence ID.'))

        return parser

    def take_action(self, parsed_args):
        """Execute rollback_vnf_instance and output comment.

        Args:
            parsed_args ([Namespace]): arguments of CLI.
        """
        client = self.app.client_manager.tackerclient
        result = client.rollback_vnf_instance(parsed_args.vnf_lcm_op_occ_id)
        if not result:
            print((_('Rollback request for LCM operation %(id)s has been'
                     ' accepted') % {'id': parsed_args.vnf_lcm_op_occ_id}))


class CancelVnfLcmOp(command.ShowOne):
    _description = _("Cancel VNF Instance")

    def get_parser(self, prog_name):
        """Add arguments to parser.

        Args:
            prog_name ([type]): program name

        Returns:
            parser([ArgumentParser]):
        """
        parser = super(CancelVnfLcmOp, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_LCM_OP_OCC_ID,
            metavar="<vnf-lcm-op-occ-id>",
            help=_('VNF lifecycle management operation occurrence ID.'))
        parser.add_argument(
            "--cancel-mode",
            default='GRACEFUL',
            metavar="<cancel-mode>",
            choices=['GRACEFUL', 'FORCEFUL'],
            help=_("Cancel mode can be 'GRACEFUL' or 'FORCEFUL'. "
                   "Default is 'GRACEFUL'"))
        return parser

    def take_action(self, parsed_args):
        """Execute cancel_vnf_instance and output comment.

        Args:
            parsed_args ([Namespace]): arguments of CLI.
        """
        client = self.app.client_manager.tackerclient
        result = client.cancel_vnf_instance(
            parsed_args.vnf_lcm_op_occ_id,
            {'cancelMode': parsed_args.cancel_mode})
        if not result:
            print((_('Cancel request for LCM operation %(id)s has been'
                     ' accepted') % {'id': parsed_args.vnf_lcm_op_occ_id}))


class FailVnfLcmOp(command.ShowOne):
    _description = _("Fail VNF Instance")

    def get_parser(self, prog_name):
        """Add arguments to parser.

        Args:
            prog_name ([type]): program name

        Returns:
            parser([ArgumentParser]):
        """
        parser = super(FailVnfLcmOp, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_LCM_OP_OCC_ID,
            metavar="<vnf-lcm-op-occ-id>",
            help=_('VNF lifecycle management operation occurrence ID.'))
        return parser

    def take_action(self, parsed_args):
        """Execute fail_vnf_instance and output response.

        Args:
            parsed_args ([Namespace]): arguments of CLI.
        """
        client = self.app.client_manager.tackerclient
        obj = client.fail_vnf_instance(parsed_args.vnf_lcm_op_occ_id)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(
            sdk_utils.DictModel(obj),
            columns, formatters=_FORMATTERS,
            mixed_case_fields=_MIXED_CASE_FIELDS)
        return (display_columns, data)


class RetryVnfLcmOp(command.Command):
    _description = _("Retry VNF Instance")

    def get_parser(self, prog_name):
        """Add arguments to parser.

        Args:
            prog_name ([type]): program name

        Returns:
            parser([ArgumentParser]):
        """

        parser = super(RetryVnfLcmOp, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_LCM_OP_OCC_ID,
            metavar="<vnf-lcm-op-occ-id>",
            help=_('VNF lifecycle management operation occurrence ID.'))
        return parser

    def take_action(self, parsed_args):
        """Execute retry_vnf_instance and output comment.

        Args:
            parsed_args ([Namespace]): arguments of CLI.
        """
        client = self.app.client_manager.tackerclient
        result = client.retry_vnf_instance(parsed_args.vnf_lcm_op_occ_id)
        if not result:
            print((_('Retry request for LCM operation %(id)s has been'
                     ' accepted') % {'id': parsed_args.vnf_lcm_op_occ_id}))


class ListVnfLcmOp(command.Lister):
    _description = _("List LCM Operation Occurrences")

    def get_parser(self, program_name):
        """Add arguments to parser.

        Args:
            program_name ([type]): program name

        Returns:
            parser([ArgumentParser]):
        """
        parser = super(ListVnfLcmOp, self).get_parser(program_name)
        parser.add_argument(
            "--filter",
            metavar="<filter>",
            help=_("Attribute-based-filtering parameters"),
        )
        fields_exclusive_group = parser.add_mutually_exclusive_group(
            required=False)
        fields_exclusive_group.add_argument(
            "--fields",
            metavar="<fields>",
            help=_("Complex attributes to be included into the response"),
        )
        fields_exclusive_group.add_argument(
            "--exclude-fields",
            metavar="<exclude-fields>",
            help=_("Complex attributes to be excluded from the response"),
        )
        return parser

    def get_attributes(self, exclude=None):
        """Get attributes.

        Args:
            exclude([exclude]): a list of fields which needs to exclude.

        Returns:
            attributes([attributes]): a list of table entry definitions.
            Each entry should be a tuple consisting of
            (API attribute name, header name, listing mode).
        """
        fields = [
            {
                "key": "id",
                "value": "ID"
            },
            {
                "key": "operationState",
                "value": "Operation State"
            },
            {
                "key": "vnfInstanceId",
                "value": "VNF Instance ID"
            },
            {
                "key": "operation",
                "value": "Operation"
            }
        ]

        attributes = []
        if exclude is None:
            exclude = []

        for field in fields:
            if field['value'] not in exclude:
                attributes.extend([(field['key'], field['value'],
                                  tacker_osc_utils.LIST_BOTH)])
        return tuple(attributes)

    def take_action(self, parsed_args):
        """Execute list_vnflcm_op_occs and output response.

        Args:
            parsed_args ([Namespace]): arguments of CLI.
        """
        params = {}
        exclude_fields = []
        extra_fields = []

        if parsed_args.filter:
            params['filter'] = parsed_args.filter
        if parsed_args.fields:
            params['fields'] = parsed_args.fields
            fields = parsed_args.fields.split(',')
            for field in fields:
                extra_fields.append(field.split('/')[0])
        if parsed_args.exclude_fields:
            params['exclude-fields'] = parsed_args.exclude_fields
            fields = parsed_args.exclude_fields.split(',')
            exclude_fields.extend(fields)

        client = self.app.client_manager.tackerclient
        vnflcm_op_occs = client.list_vnf_lcm_op_occs(**params)
        headers, columns = tacker_osc_utils.get_column_definitions(
            self.get_attributes(exclude=exclude_fields),
            long_listing=True)

        dictionary_properties = (utils.get_dict_properties(
            s, columns, mixed_case_fields=_MIXED_CASE_FIELDS)
            for s in vnflcm_op_occs
        )

        return (headers, dictionary_properties)


class ShowVnfLcmOp(command.ShowOne):
    _description = _("Display Operation Occurrence details")

    def get_parser(self, program_name):
        """Add arguments to parser.

        Args:
            program_name ([type]): program name

        Returns:
            parser([ArgumentParser]):
        """
        parser = super(ShowVnfLcmOp, self).get_parser(program_name)
        parser.add_argument(
            _VNF_LCM_OP_OCC_ID,
            metavar="<vnf-lcm-op-occ-id>",
            help=_('VNF lifecycle management operation occurrence ID.'))
        return parser

    def take_action(self, parsed_args):
        """Execute show_vnf_lcm_op_occs and output response.

        Args:
            parsed_args ([Namespace]): arguments of CLI.
        """
        client = self.app.client_manager.tackerclient
        obj = client.show_vnf_lcm_op_occs(parsed_args.vnf_lcm_op_occ_id)
        display_columns, columns = _get_columns(obj, action='show')
        data = utils.get_item_properties(
            sdk_utils.DictModel(obj),
            columns, formatters=_FORMATTERS,
            mixed_case_fields=_MIXED_CASE_FIELDS)
        return (display_columns, data)



# Copyright (C) 2013 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools
from unittest import mock

from tackerclient.client import HTTPClient
from tackerclient.common import exceptions
from tackerclient.tests.unit.test_cli10 import MyResp


AUTH_TOKEN = 'test_token'
END_URL = 'test_url'
METHOD = 'GET'
URL = 'http://test.test:1234/v1.0/test'
headers = {'User-Agent': 'python-tackerclient'}


class TestHTTPClient(testtools.TestCase):

    def setUp(self):

        super(TestHTTPClient, self).setUp()
        self.addCleanup(mock.patch.stopall)
        self.http = HTTPClient(token=AUTH_TOKEN, endpoint_url=END_URL)

    @mock.patch('tackerclient.client.HTTPClient.request')
    def test_request_error(self, mock_request):

        mock_request.side_effect = Exception('error msg')
        self.assertRaises(
            exceptions.ConnectionFailed,
            self.http._cs_request,
            URL, METHOD
        )

    @mock.patch('tackerclient.client.HTTPClient.request')
    def test_request_success(self, mock_request):

        rv_should_be = MyResp(200), 'test content'
        mock_request.return_value = rv_should_be
        self.assertEqual(rv_should_be, self.http._cs_request(URL, METHOD))

    @mock.patch('tackerclient.client.HTTPClient.request')
    def test_request_unauthorized(self, mock_request):

        mock_request.return_value = MyResp(401), 'unauthorized message'

        e = self.assertRaises(exceptions.Unauthorized,
                              self.http._cs_request, URL, METHOD)
        self.assertEqual('unauthorized message', str(e))
        mock_request.assert_called_with(URL, METHOD, headers=headers)

    @mock.patch('tackerclient.client.HTTPClient.request')
    def test_request_forbidden_is_returned_to_caller(self, mock_request):

        rv_should_be = MyResp(403), 'forbidden message'
        mock_request.return_value = rv_should_be
        self.assertEqual(rv_should_be, self.http._cs_request(URL, METHOD))



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import testtools

from tackerclient.common import exceptions
from tackerclient.tacker import v1_0 as tackerV10


class CLITestArgs(testtools.TestCase):

    def test_empty(self):
        _mydict = tackerV10.parse_args_to_dict([])
        self.assertEqual({}, _mydict)

    def test_default_bool(self):
        _specs = ['--my_bool', '--arg1', 'value1']
        _mydict = tackerV10.parse_args_to_dict(_specs)
        self.assertTrue(_mydict['my_bool'])

    def test_bool_true(self):
        _specs = ['--my-bool', 'type=bool', 'true', '--arg1', 'value1']
        _mydict = tackerV10.parse_args_to_dict(_specs)
        self.assertTrue(_mydict['my_bool'])

    def test_bool_false(self):
        _specs = ['--my_bool', 'type=bool', 'false', '--arg1', 'value1']
        _mydict = tackerV10.parse_args_to_dict(_specs)
        self.assertFalse(_mydict['my_bool'])

    def test_nargs(self):
        _specs = ['--tag', 'x', 'y', '--arg1', 'value1']
        _mydict = tackerV10.parse_args_to_dict(_specs)
        self.assertIn('x', _mydict['tag'])
        self.assertIn('y', _mydict['tag'])

    def test_badarg(self):
        _specs = ['--tag=t', 'x', 'y', '--arg1', 'value1']
        self.assertRaises(exceptions.CommandError,
                          tackerV10.parse_args_to_dict, _specs)

    def test_badarg_with_minus(self):
        _specs = ['--arg1', 'value1', '-D']
        self.assertRaises(exceptions.CommandError,
                          tackerV10.parse_args_to_dict, _specs)

    def test_goodarg_with_minus_number(self):
        _specs = ['--arg1', 'value1', '-1', '-1.0']
        _mydict = tackerV10.parse_args_to_dict(_specs)
        self.assertEqual(['value1', '-1', '-1.0'],
                         _mydict['arg1'])

    def test_badarg_duplicate(self):
        _specs = ['--tag=t', '--arg1', 'value1', '--arg1', 'value1']
        self.assertRaises(exceptions.CommandError,
                          tackerV10.parse_args_to_dict, _specs)

    def test_badarg_early_type_specification(self):
        _specs = ['type=dict', 'key=value']
        self.assertRaises(exceptions.CommandError,
                          tackerV10.parse_args_to_dict, _specs)

    def test_arg(self):
        _specs = ['--tag=t', '--arg1', 'value1']
        self.assertEqual('value1',
                         tackerV10.parse_args_to_dict(_specs)['arg1'])

    def test_dict_arg(self):
        _specs = ['--tag=t', '--arg1', 'type=dict', 'key1=value1,key2=value2']
        arg1 = tackerV10.parse_args_to_dict(_specs)['arg1']
        self.assertEqual('value1', arg1['key1'])
        self.assertEqual('value2', arg1['key2'])

    def test_dict_arg_with_attribute_named_type(self):
        _specs = ['--tag=t', '--arg1', 'type=dict', 'type=value1,key2=value2']
        arg1 = tackerV10.parse_args_to_dict(_specs)['arg1']
        self.assertEqual('value1', arg1['type'])
        self.assertEqual('value2', arg1['key2'])

    def test_list_of_dict_arg(self):
        _specs = ['--tag=t', '--arg1', 'type=dict',
                  'list=true', 'key1=value1,key2=value2']
        arg1 = tackerV10.parse_args_to_dict(_specs)['arg1']
        self.assertEqual('value1', arg1[0]['key1'])
        self.assertEqual('value2', arg1[0]['key2'])

    def test_clear_action(self):
        _specs = ['--anyarg', 'action=clear']
        args = tackerV10.parse_args_to_dict(_specs)
        self.assertIsNone(args['anyarg'])

    def test_bad_values_str(self):
        _specs = ['--strarg', 'type=str']
        self.assertRaises(exceptions.CommandError,
                          tackerV10.parse_args_to_dict, _specs)

    def test_bad_values_list(self):
        _specs = ['--listarg', 'list=true', 'type=str']
        self.assertRaises(exceptions.CommandError,
                          tackerV10.parse_args_to_dict, _specs)
        _specs = ['--listarg', 'type=list']
        self.assertRaises(exceptions.CommandError,
                          tackerV10.parse_args_to_dict, _specs)
        _specs = ['--listarg', 'type=list', 'action=clear']
        self.assertRaises(exceptions.CommandError,
                          tackerV10.parse_args_to_dict, _specs)



# Copyright (C) 2022 Fujitsu
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import os
import sys

from io import StringIO
from oslo_utils.fixture import uuidsentinel
from unittest import mock

from tackerclient import client as root_client
from tackerclient.common import exceptions
from tackerclient.osc import utils as tacker_osc_utils
from tackerclient.osc.v2.vnfpm import vnfpm_job
from tackerclient.tests.unit.osc import base
from tackerclient.tests.unit.osc.v1.fixture_data import client
from tackerclient.tests.unit.osc.v2 import vnfpm_job_fakes
from tackerclient.tests.unit.test_cli10 import MyResp
from tackerclient.v1_0 import client as proxy_client


class TestVnfPmJob(base.FixturedTestCase):
    client_fixture_class = client.ClientFixture

    def setUp(self):
        super(TestVnfPmJob, self).setUp()
        self.url = client.TACKER_URL
        self.header = {'content-type': 'application/json'}
        self.app = mock.Mock()
        self.app_args = mock.Mock()
        self.client_manager = self.cs
        self.app.client_manager.tackerclient = self.client_manager


def _get_columns_vnfpm_job(action=None):
    if action == 'update':
        columns = ['Callback Uri']
    else:
        columns = ['ID', 'Object Type', 'Object Instance Ids',
                   'Sub Object Instance Ids', 'Criteria', 'Callback Uri',
                   'Reports', 'Links']
    if action == 'list':
        columns = [
            ele for ele in columns if ele not in [
                'Criteria', 'Sub Object Instance Ids', 'Reports', 'Links'
            ]
        ]
    return columns


@ddt.ddt
class TestCreateVnfPmJob(TestVnfPmJob):

    def setUp(self):
        super(TestCreateVnfPmJob, self).setUp()
        self.create_vnf_pm_job = vnfpm_job.CreateVnfPmJob(
            self.app, self.app_args, cmd_name='vnfpm job create')

    def test_create_no_args(self):
        self.assertRaises(base.ParserException, self.check_parser,
                          self.create_vnf_pm_job, [], [])

    @ddt.unpack
    def test_take_action(self):
        param_file = ("./tackerclient/osc/v2/vnfpm/samples/"
                      "create_vnf_pm_job_param_sample.json")

        arg_list = [param_file]
        verify_list = [('request_file', param_file)]

        parsed_args = self.check_parser(self.create_vnf_pm_job, arg_list,
                                        verify_list)

        json = vnfpm_job_fakes.vnf_pm_job_response()
        self.requests_mock.register_uri(
            'POST', os.path.join(self.url, 'vnfpm/v2/pm_jobs'),
            json=json, headers=self.header)

        actual_columns, data = (
            self.create_vnf_pm_job.take_action(parsed_args))
        self.assertCountEqual(_get_columns_vnfpm_job(),
                              actual_columns)

        _, attributes = vnfpm_job._get_columns(json)
        expected_data = vnfpm_job_fakes.get_vnfpm_job_data(
            json, columns=attributes)
        self.assertListItemsEqual(expected_data, data)


@ddt.ddt
class TestListVnfPmJob(TestVnfPmJob):

    def setUp(self):
        super(TestListVnfPmJob, self).setUp()
        self.list_vnf_pm_jobs = vnfpm_job.ListVnfPmJob(
            self.app, self.app_args, cmd_name='vnfpm job list')
        self._vnf_pm_jobs = self._get_vnf_pm_jobs()

    def _get_vnf_pm_jobs(self):
        return vnfpm_job_fakes.create_vnf_pm_jobs(count=3)

    def get_list_columns(self, all_fields=False, exclude_fields=None,
                         extra_fields=None, exclude_default=False):

        columns = ['Id', 'Object Type', 'Links']
        complex_columns = [
            'Object Instance Ids',
            'Sub Object Instance Ids',
            'Criteria',
            'Reports'
        ]
        simple_columns = ['Callback Uri']

        if extra_fields:
            columns.extend(extra_fields)

        if exclude_fields:
            columns.extend([field for field in complex_columns
                           if field not in exclude_fields])
        if all_fields:
            columns.extend(complex_columns)
            columns.extend(simple_columns)

        if exclude_default:
            columns.extend(simple_columns)

        return columns

    def _get_mock_response_for_list_vnf_pm_jobs(
            self, filter_attribute, json=None):
        self.requests_mock.register_uri(
            'GET', self.url + '/vnfpm/v2/pm_jobs?' + filter_attribute,
            json=json if json else self._get_vnf_pm_jobs(),
            headers=self.header)

    def test_take_action_default_fields(self):
        parsed_args = self.check_parser(self.list_vnf_pm_jobs, [], [])
        self.requests_mock.register_uri(
            'GET', self.url + '/vnfpm/v2/pm_jobs',
            json=self._vnf_pm_jobs, headers=self.header)
        actual_columns, data = self.list_vnf_pm_jobs.take_action(parsed_args)
        expected_data = []
        _, columns = tacker_osc_utils.get_column_definitions(
            self.list_vnf_pm_jobs.get_attributes(), long_listing=True)

        for vnf_pm_job_obj in self._vnf_pm_jobs['vnf_pm_jobs']:
            expected_data.append(vnfpm_job_fakes.get_vnfpm_job_data(
                vnf_pm_job_obj, columns=columns))
        self.assertCountEqual(self.get_list_columns(), actual_columns)
        self.assertListItemsEqual(expected_data, list(data))

    @ddt.data('all_fields', 'exclude_default')
    def test_take_action(self, arg):
        parsed_args = self.check_parser(
            self.list_vnf_pm_jobs,
            ["--" + arg, "--filter", '(eq,objectType,VNFC)'],
            [(arg, True), ('filter', '(eq,objectType,VNFC)')])
        vnf_pm_jobs = self._get_vnf_pm_jobs()
        self._get_mock_response_for_list_vnf_pm_jobs(
            'filter=(eq,objectType,VNFC)&' + arg, json=vnf_pm_jobs)

        actual_columns, data = self.list_vnf_pm_jobs.take_action(parsed_args)
        expected_data = []
        kwargs = {arg: True}
        _, columns = tacker_osc_utils.get_column_definitions(
            self.list_vnf_pm_jobs.get_attributes(**kwargs), long_listing=True)

        for vnf_pm_job_obj in vnf_pm_jobs['vnf_pm_jobs']:
            expected_data.append(vnfpm_job_fakes.get_vnfpm_job_data(
                vnf_pm_job_obj, columns=columns))

        self.assertCountEqual(self.get_list_columns(**kwargs), actual_columns)
        self.assertListItemsEqual(expected_data, list(data))

    def test_take_action_with_exclude_fields(self):
        parsed_args = self.check_parser(
            self.list_vnf_pm_jobs,
            ["--exclude_fields", 'objectInstanceIds,criteria',
             "--filter", '(eq,objectType,VNFC)'],
            [('exclude_fields', 'objectInstanceIds,criteria'),
             ('filter', '(eq,objectType,VNFC)')])
        vnf_pm_jobs = self._get_vnf_pm_jobs()
        updated_vnf_pm_jobs = {'vnf_pm_jobs': []}
        for vnf_pm_job_obj in vnf_pm_jobs['vnf_pm_jobs']:
            vnf_pm_job_obj.pop('objectInstanceIds')
            vnf_pm_job_obj.pop('criteria')
            updated_vnf_pm_jobs['vnf_pm_jobs'].append(vnf_pm_job_obj)
        self._get_mock_response_for_list_vnf_pm_jobs(
            'filter=(eq,objectType,VNFC)&'
            'exclude_fields=objectInstanceIds,criteria',
            json=updated_vnf_pm_jobs)

        actual_columns, data = self.list_vnf_pm_jobs.take_action(parsed_args)
        expected_data = []
        _, columns = tacker_osc_utils.get_column_definitions(
            self.list_vnf_pm_jobs.get_attributes(
                exclude_fields=['objectInstanceIds', 'criteria']),
            long_listing=True)

        for updated_vnf_pm_obj in updated_vnf_pm_jobs['vnf_pm_jobs']:
            expected_data.append(vnfpm_job_fakes.get_vnfpm_job_data(
                updated_vnf_pm_obj, columns=columns))
        expected_columns = self.get_list_columns(
            exclude_fields=['Object Instance Ids', 'Criteria'])
        self.assertCountEqual(expected_columns, actual_columns)
        self.assertListItemsEqual(expected_data, list(data))

    @ddt.data((['--all_fields', '--fields', 'objectInstanceIds'],
               [('all_fields', True), ('fields', 'objectInstanceIds')]),
              (['--all_fields', '--exclude_fields', 'criteria'],
               [('all_fields', True), ('exclude_fields', 'criteria')]),
              (['--fields', 'objectInstanceIds',
                '--exclude_fields', 'criteria'],
               [('fields', 'objectInstanceIds'),
                ('exclude_fields', 'criteria')]))
    @ddt.unpack
    def test_take_action_with_invalid_combination(self, arglist, verifylist):
        self.assertRaises(base.ParserException, self.check_parser,
                          self.list_vnf_pm_jobs, arglist, verifylist)

    def test_take_action_with_valid_combination(self):
        parsed_args = self.check_parser(
            self.list_vnf_pm_jobs,
            [
                "--fields", 'subObjectInstanceIds,criteria',
                "--exclude_default"
            ],
            [
                ('fields', 'subObjectInstanceIds,criteria'),
                ('exclude_default', True)
            ])
        vnf_pm_jobs = self._get_vnf_pm_jobs()

        updated_vnf_pm_jobs = {'vnf_pm_jobs': []}
        for vnf_pm_job_obj in vnf_pm_jobs['vnf_pm_jobs']:
            # vnf_pm_job_obj.pop('userDefinedData')
            updated_vnf_pm_jobs['vnf_pm_jobs'].append(vnf_pm_job_obj)

        self._get_mock_response_for_list_vnf_pm_jobs(
            'exclude_default&fields=subObjectInstanceIds,criteria',
            json=updated_vnf_pm_jobs)

        actual_columns, data = self.list_vnf_pm_jobs.take_action(parsed_args)
        expected_data = []
        _, columns = tacker_osc_utils.get_column_definitions(
            self.list_vnf_pm_jobs.get_attributes(
                extra_fields=['subObjectInstanceIds', 'criteria'],
                exclude_default=True),
            long_listing=True)

        for updated_vnf_pm_job_obj in updated_vnf_pm_jobs['vnf_pm_jobs']:
            expected_data.append(vnfpm_job_fakes.get_vnfpm_job_data(
                updated_vnf_pm_job_obj, columns=columns))

        expected_columns = self.get_list_columns(
            extra_fields=['Sub Object Instance Ids', 'Criteria'],
            exclude_default=True)
        self.assertCountEqual(expected_columns, actual_columns)
        self.assertListItemsEqual(expected_data, list(data))


class TestShowVnfPmJob(TestVnfPmJob):

    def setUp(self):
        super(TestShowVnfPmJob, self).setUp()
        self.show_vnf_pm_jobs = vnfpm_job.ShowVnfPmJob(
            self.app, self.app_args, cmd_name='vnfpm job show')

    def test_take_action(self):
        """Test of take_action()"""
        vnfpm_job_obj = vnfpm_job_fakes.vnf_pm_job_response()

        arg_list = [vnfpm_job_obj['id']]
        verify_list = [('vnf_pm_job_id', vnfpm_job_obj['id'])]

        # command param
        parsed_args = self.check_parser(
            self.show_vnf_pm_jobs, arg_list, verify_list)
        url = os.path.join(
            self.url, 'vnfpm/v2/pm_jobs', vnfpm_job_obj['id'])

        self.requests_mock.register_uri(
            'GET', url, headers=self.header, json=vnfpm_job_obj)

        columns, data = (self.show_vnf_pm_jobs.take_action(parsed_args))

        self.assertCountEqual(_get_columns_vnfpm_job('show'), columns)

        _, attributes = vnfpm_job._get_columns(vnfpm_job_obj)
        self.assertListItemsEqual(
            vnfpm_job_fakes.get_vnfpm_job_data(
                vnfpm_job_obj, columns=attributes), data)

    def test_take_action_vnf_pm_job_id_not_found(self):
        """Test if vnf-pm-job-id does not find."""
        arg_list = [uuidsentinel.vnf_pm_job_id]
        verify_list = [('vnf_pm_job_id', uuidsentinel.vnf_pm_job_id)]

        # command param
        parsed_args = self.check_parser(
            self.show_vnf_pm_jobs, arg_list, verify_list)

        url = os.path.join(
            self.url, 'vnfpm/v2/pm_jobs', uuidsentinel.vnf_pm_job_id)
        self.requests_mock.register_uri(
            'GET', url, headers=self.header, status_code=404, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.show_vnf_pm_jobs.take_action,
                          parsed_args)

    def test_take_action_internal_server_error(self):
        """Test for internal server error."""
        arg_list = [uuidsentinel.vnf_pm_job_id]
        verify_list = [('vnf_pm_job_id', uuidsentinel.vnf_pm_job_id)]

        # command param
        parsed_args = self.check_parser(
            self.show_vnf_pm_jobs, arg_list, verify_list)

        url = os.path.join(
            self.url, 'vnfpm/v2/pm_jobs', uuidsentinel.vnf_pm_job_id)
        self.requests_mock.register_uri(
            'GET', url, headers=self.header, status_code=500, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.show_vnf_pm_jobs.take_action,
                          parsed_args)


@ddt.ddt
class TestUpdateVnfPmJob(TestVnfPmJob):

    def setUp(self):
        super(TestUpdateVnfPmJob, self).setUp()
        self.update_vnf_pm_job = vnfpm_job.UpdateVnfPmJob(
            self.app, self.app_args, cmd_name='vnfpm job update')

    def test_take_action(self):
        """Test of take_action()"""

        param_file = ("./tackerclient/osc/v2/vnfpm/samples/"
                      "update_vnf_pm_job_param_sample.json")
        arg_list = [uuidsentinel.vnf_pm_job_id, param_file]
        verify_list = [
            ('vnf_pm_job_id', uuidsentinel.vnf_pm_job_id),
            ('request_file', param_file)
        ]
        vnfpm_job_obj = vnfpm_job_fakes.vnf_pm_job_response(
            None, 'update')

        # command param
        parsed_args = self.check_parser(
            self.update_vnf_pm_job, arg_list, verify_list)
        url = os.path.join(
            self.url, 'vnfpm/v2/pm_jobs', uuidsentinel.vnf_pm_job_id)
        self.requests_mock.register_uri(
            'PATCH', url, headers=self.header, json=vnfpm_job_obj)

        actual_columns, data = (
            self.update_vnf_pm_job.take_action(parsed_args))
        expected_columns = _get_columns_vnfpm_job(action='update')
        self.assertCountEqual(expected_columns, actual_columns)

        _, columns = vnfpm_job._get_columns(
            vnfpm_job_obj, action='update')
        expected_data = vnfpm_job_fakes.get_vnfpm_job_data(
            vnfpm_job_obj, columns=columns)
        self.assertEqual(expected_data, data)

    @mock.patch.object(proxy_client.ClientBase, 'deserialize')
    def test_take_action_check_content_type(self, mock_des):
        """Check content type by test of take_action()"""

        param_file = ('./tackerclient/osc/v2/vnfpm/samples/'
                      'update_vnf_pm_job_param_sample.json')
        arg_list = [uuidsentinel.vnf_pm_job_id, param_file]
        verify_list = [
            ('vnf_pm_job_id', uuidsentinel.vnf_pm_job_id),
            ('request_file', param_file)
        ]
        vnfpm_job_obj = vnfpm_job_fakes.vnf_pm_job_response(
            None, 'update')
        mock_des.return_value = {}

        # command param
        parsed_args = self.check_parser(
            self.update_vnf_pm_job, arg_list, verify_list)
        url = os.path.join(
            self.url, 'vnfpm/v2/pm_jobs', uuidsentinel.vnf_pm_job_id)
        self.requests_mock.register_uri(
            'PATCH', url, headers=self.header, json=vnfpm_job_obj)

        with mock.patch.object(root_client.HTTPClient,
                               'do_request') as mock_req:
            headers = {'Content-Type': 'application/json'}
            mock_req.return_value = (MyResp(200, headers=headers), None)
            self.update_vnf_pm_job.take_action(parsed_args)
            mock_req.assert_called_once_with(
                f'/vnfpm/v2/pm_jobs/{uuidsentinel.vnf_pm_job_id}', 'PATCH',
                body=mock.ANY, headers=mock.ANY,
                content_type='application/merge-patch+json', accept='json')

    def test_take_action_vnf_pm_job_id_not_found(self):
        """Test if vnf-pm-job-id does not find"""

        param_file = ("./tackerclient/osc/v2/vnfpm/samples/"
                      "update_vnf_pm_job_param_sample.json")
        arg_list = [uuidsentinel.vnf_pm_job_id, param_file]
        verify_list = [
            ('vnf_pm_job_id', uuidsentinel.vnf_pm_job_id),
            ('request_file', param_file)
        ]

        # command param
        parsed_args = self.check_parser(
            self.update_vnf_pm_job, arg_list, verify_list)
        url = os.path.join(
            self.url, 'vnfpm/v2/pm_jobs', uuidsentinel.vnf_pm_job_id)
        self.requests_mock.register_uri(
            'PATCH', url, headers=self.header, status_code=404, json={})
        self.assertRaises(exceptions.TackerClientException,
                          self.update_vnf_pm_job.take_action,
                          parsed_args)


class TestDeleteVnfPmJob(TestVnfPmJob):

    def setUp(self):
        super(TestDeleteVnfPmJob, self).setUp()
        self.delete_vnf_pm_job = vnfpm_job.DeleteVnfPmJob(
            self.app, self.app_args, cmd_name='vnfpm job delete')

        # Vnf Fm job to delete
        self.vnf_pm_jobs = vnfpm_job_fakes.create_vnf_pm_jobs(
            count=3)['vnf_pm_jobs']

    def _mock_request_url_for_delete(self, index):
        url = os.path.join(self.url, 'vnfpm/v2/pm_jobs',
                           self.vnf_pm_jobs[index]['id'])

        self.requests_mock.register_uri('DELETE', url,
                                        headers=self.header, json={})

    def test_delete_one_vnf_pm_job(self):
        arg_list = [self.vnf_pm_jobs[0]['id']]
        verify_list = [('vnf_pm_job_id',
                       [self.vnf_pm_jobs[0]['id']])]

        parsed_args = self.check_parser(self.delete_vnf_pm_job, arg_list,
                                        verify_list)

        self._mock_request_url_for_delete(0)
        sys.stdout = buffer = StringIO()
        result = self.delete_vnf_pm_job.take_action(parsed_args)
        self.assertIsNone(result)
        self.assertEqual(
            (f"VNF PM job '{self.vnf_pm_jobs[0]['id']}' "
             f"deleted successfully"), buffer.getvalue().strip())

    def test_delete_multiple_vnf_pm_job(self):
        arg_list = []
        for obj in self.vnf_pm_jobs:
            arg_list.append(obj['id'])
        verify_list = [('vnf_pm_job_id', arg_list)]
        parsed_args = self.check_parser(self.delete_vnf_pm_job, arg_list,
                                        verify_list)
        for i in range(0, 3):
            self._mock_request_url_for_delete(i)
        sys.stdout = buffer = StringIO()
        result = self.delete_vnf_pm_job.take_action(parsed_args)
        self.assertIsNone(result)
        self.assertEqual('All specified VNF PM jobs are deleted '
                         'successfully', buffer.getvalue().strip())

    def test_delete_multiple_vnf_pm_job_exception(self):
        arg_list = [
            self.vnf_pm_jobs[0]['id'],
            'xxxx-yyyy-zzzz',
            self.vnf_pm_jobs[1]['id'],
        ]
        verify_list = [('vnf_pm_job_id', arg_list)]
        parsed_args = self.check_parser(self.delete_vnf_pm_job,
                                        arg_list, verify_list)

        self._mock_request_url_for_delete(0)

        url = os.path.join(self.url, 'vnfpm/v2/jobs',
                           'xxxx-yyyy-zzzz')
        self.requests_mock.register_uri(
            'GET', url, exc=exceptions.ConnectionFailed)

        self._mock_request_url_for_delete(1)
        exception = self.assertRaises(exceptions.CommandError,
                                      self.delete_vnf_pm_job.take_action,
                                      parsed_args)

        self.assertEqual('Failed to delete 1 of 3 VNF PM jobs.',
                         exception.message)



# Copyright 2016 Brocade Communications Systems Inc
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaml

from osc_lib.command import command
from osc_lib import utils
from oslo_utils import strutils

from tackerclient.common import exceptions
from tackerclient.i18n import _
from tackerclient.osc import sdk_utils
from tackerclient.osc import utils as tacker_osc_utils
from tackerclient.tacker import v1_0 as tackerV10
from tackerclient.tacker.v1_0.nfvo import vim_utils

_attr_map = (
    ('id', 'ID', tacker_osc_utils.LIST_BOTH),
    ('name', 'Name', tacker_osc_utils.LIST_BOTH),
    ('tenant_id', 'Tenant_id', tacker_osc_utils.LIST_BOTH),
    ('type', 'Type', tacker_osc_utils.LIST_BOTH),
    ('is_default', 'Is Default',
     tacker_osc_utils.LIST_BOTH),
    ('placement_attr', 'Placement attribution',
     tacker_osc_utils.LIST_LONG_ONLY),
    ('status', 'Status', tacker_osc_utils.LIST_BOTH),
)

_VIM = 'vim'


class ListVIM(command.Lister):
    _description = _("List VIMs that belong to a given tenant.")

    def get_parser(self, prog_name):
        parser = super(ListVIM, self).get_parser(prog_name)
        parser.add_argument(
            '--long',
            action='store_true',
            help=_("List additional fields in output")
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        data = client.list_vims()
        headers, columns = tacker_osc_utils.get_column_definitions(
            _attr_map, long_listing=parsed_args.long)
        return (headers,
                (utils.get_dict_properties(
                    s, columns,
                ) for s in data[_VIM + 's']))


class ShowVIM(command.ShowOne):
    _description = _("Display VIM details")

    def get_parser(self, prog_name):
        parser = super(ShowVIM, self).get_parser(prog_name)
        parser.add_argument(
            _VIM,
            metavar="<VIM>",
            help=_("VIM to display (name or ID)")
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        obj_id = tackerV10.find_resourceid_by_name_or_id(
            client, _VIM, parsed_args.vim)
        obj = client.show_vim(obj_id)
        display_columns, columns = _get_columns(obj[_VIM])
        data = utils.get_item_properties(
            sdk_utils.DictModel(obj[_VIM]),
            columns,
            formatters=_formatters)
        return (display_columns, data)


class CreateVIM(command.ShowOne):
    _description = _("Register a new VIM")

    def get_parser(self, prog_name):
        parser = super(CreateVIM, self).get_parser(prog_name)
        parser.add_argument(
            'name', metavar='NAME',
            help=_('Set a name for the VIM'))
        parser.add_argument(
            '--tenant-id', metavar='TENANT_ID',
            help=_('The owner tenant ID or project ID'))
        parser.add_argument(
            '--config-file',
            required=True,
            help=_('YAML file with VIM configuration parameters'))
        parser.add_argument(
            '--description',
            help=_('Set a description for the VIM'))
        parser.add_argument(
            '--is-default',
            action='store_true',
            default=False,
            help=_('Set as default VIM'))
        return parser

    def args2body(self, parsed_args):
        body = {_VIM: {}}
        if parsed_args.config_file:
            with open(parsed_args.config_file) as f:
                vim_config = f.read()
                try:
                    config_param = yaml.load(vim_config,
                                             Loader=yaml.SafeLoader)
                except yaml.YAMLError as e:
                    raise exceptions.InvalidInput(reason=e)
        vim_obj = body[_VIM]
        try:
            auth_url = config_param.pop('auth_url')
        except KeyError:
            raise exceptions.TackerClientException(message='Auth URL must be '
                                                           'specified',
                                                   status_code=404)
        vim_obj['auth_url'] = vim_utils.validate_auth_url(auth_url).geturl()
        vim_utils.args2body_vim(config_param, vim_obj)
        tackerV10.update_dict(parsed_args, body[_VIM],
                              ['tenant_id', 'name', 'description',
                               'is_default'])
        return body

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        vim = client.create_vim(self.args2body(parsed_args))
        display_columns, columns = _get_columns(vim[_VIM])
        data = utils.get_item_properties(
            sdk_utils.DictModel(vim[_VIM]),
            columns, formatters=_formatters)
        return (display_columns, data)


class DeleteVIM(command.Command):
    _description = _("Delete VIM(s).")

    def get_parser(self, prog_name):
        parser = super(DeleteVIM, self).get_parser(prog_name)
        parser.add_argument(
            _VIM,
            metavar="<VIM>",
            nargs="+",
            help=_("VIM(s) to delete (name or ID)")
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        failure = False
        deleted_ids = []
        failed_items = {}
        for resource_id in parsed_args.vim:
            try:
                obj = tackerV10.find_resourceid_by_name_or_id(
                    client, _VIM, resource_id)
                client.delete_vim(obj)
                deleted_ids.append(resource_id)
            except Exception as e:
                failure = True
                failed_items[resource_id] = e
        if failure:
            msg = ''
            if deleted_ids:
                msg = (_('Successfully deleted %(resource)s(s):'
                         ' %(deleted_list)s') % {'deleted_list':
                                                 ', '.join(deleted_ids),
                                                 'resource': _VIM})
            err_msg = _("\n\nUnable to delete the below"
                        " %s(s):") % _VIM
            for failed_id, error in failed_items.items():
                err_msg += (_('\n Cannot delete %(failed_id)s: %(error)s')
                            % {'failed_id': failed_id,
                               'error': error})
            msg += err_msg
            raise exceptions.CommandError(message=msg)
        else:
            print((_('All specified %(resource)s(s) deleted successfully')
                   % {'resource': _VIM}))
        return


class UpdateVIM(command.ShowOne):
    _description = _("Update VIM.")

    def get_parser(self, prog_name):
        parser = super(UpdateVIM, self).get_parser(prog_name)
        parser.add_argument(
            'id', metavar="VIM",
            help=_('ID or name of %s to update') % _VIM)
        parser.add_argument(
            '--config-file',
            required=False,
            help=_('YAML file with VIM configuration parameters'))
        parser.add_argument(
            '--name',
            help=_('New name for the VIM'))
        parser.add_argument(
            '--description',
            help=_('New description for the VIM'))
        parser.add_argument(
            '--is-default',
            type=strutils.bool_from_string,
            metavar='{True,False}',
            help=_('Indicate whether the VIM is used as default'))
        return parser

    def args2body(self, parsed_args):
        body = {_VIM: {}}
        config_param = None
        # config arg passed as data overrides config yaml when both args passed
        if parsed_args.config_file:
            with open(parsed_args.config_file) as f:
                config_yaml = f.read()
            try:
                config_param = yaml.load(config_yaml,
                                         Loader=yaml.SafeLoader)
            except yaml.YAMLError as e:
                raise exceptions.InvalidInput(reason=e)
        vim_obj = body[_VIM]
        if config_param is not None:
            vim_utils.args2body_vim(config_param, vim_obj)
        tackerV10.update_dict(parsed_args, body[_VIM],
                              ['tenant_id', 'name', 'description',
                               'is_default'])
        # type attribute is read-only, it can't be updated, so remove it
        # in update method
        body[_VIM].pop('type', None)
        return body

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        obj_id = tackerV10.find_resourceid_by_name_or_id(
            client, _VIM, parsed_args.id)
        vim = client.update_vim(obj_id, self.args2body(parsed_args))
        display_columns, columns = _get_columns(vim[_VIM])
        data = utils.get_item_properties(
            sdk_utils.DictModel(vim[_VIM]), columns,
            formatters=_formatters)
        return (display_columns, data)


_formatters = {
    'auth_cred': tacker_osc_utils.format_dict_with_indention,
    'placement_attr': tacker_osc_utils.format_dict_with_indention,
    'vim_project': tacker_osc_utils.format_dict_with_indention,
}


def _get_columns(item):
    column_map = {
        'tenant_id': 'project_id',
    }
    return sdk_utils.get_osc_show_columns_for_sdk_resource(item, column_map)



# Copyright (C) 2022 Fujitsu
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from tackerclient.osc import utils as tacker_osc_utils


def create_vnf_pm_jobs(count=2):
    """Create multiple fake vnf pm jobs.

    :param int count:
        The number of vnf_pm_jobs to fake
    :return:
        A list of fake vnf pm jobs dictionary
    """
    vnf_pm_jobs = []
    for i in range(0, count):
        unique_id = uuidutils.generate_uuid()
        vnf_pm_jobs.append(vnf_pm_job_response(attrs={'id': unique_id}))
    return {'vnf_pm_jobs': vnf_pm_jobs}


def vnf_pm_job_response(attrs=None, action=None):
    """Create a fake vnf pm job.

    :param Dictionary attrs:
        A dictionary with all attributes
    :return:
        A pm job dict
    """
    if action == 'update':
        fake_vnf_pm_job = {
            "callbackUri": "/nfvo/notify/job"
        }
        return fake_vnf_pm_job

    attrs = attrs or {}
    # Set default attributes.
    fake_vnf_pm_job = {
        "id": "2bb72d78-b1d9-48fe-8c64-332654ffeb5d",
        "objectType": "VNFC",
        "objectInstanceIds": [
            "object-instance-id-1"
        ],
        "subObjectInstanceIds": [
            "sub-object-instance-id-2"
        ],
        "criteria": {
            "performanceMetric": [
                "VCpuUsageMeanVnf.object-instance-id-1"
            ],
            "performanceMetricGroup": [
                "VirtualisedComputeResource"
            ],
            "collectionPeriod": 500,
            "reportingPeriod": 1000,
            "reportingBoundary": "2022/07/25 10:43:55"
        },
        "callbackUri": "/nfvo/notify/job",
        "reports": [{
            "href": "/vnfpm/v2/pm_jobs/2bb72d78-b1d9-48fe-8c64-332654ffeb5d/"
                    "reports/09d46aed-3ec2-45d9-bfa2-add431e069b3",
            "readyTime": "2022/07/25 10:43:55",
            "expiryTime": "2022/07/25 10:43:55",
            "fileSize": 9999
        }],
        "_links": {
            "self": {
                "href": "/vnfpm/v2/pm_jobs/"
                        "78a39661-60a8-4824-b989-88c1b0c3534a"
            },
            "objects": [{
                "href": "/vnflcm/v1/vnf_instances/"
                        "0e5f3086-4e79-47ed-a694-54c29155fa26"
            }]
        }
    }

    # Overwrite default attributes.
    fake_vnf_pm_job.update(attrs)

    return fake_vnf_pm_job


def get_vnfpm_job_data(vnf_pm_job, columns=None):
    """Get the vnfpm job.

    :return:
        A tuple object sorted based on the name of the columns.
    """
    complex_attributes = [
        'objectInstanceIds', 'subObjectInstanceIds',
        'criteria', 'reports', '_links',
        'authentication'
    ]

    for attribute in complex_attributes:
        if vnf_pm_job.get(attribute):
            vnf_pm_job.update(
                {attribute: tacker_osc_utils.FormatComplexDataColumn(
                    vnf_pm_job[attribute])})

    # return the list of data as per column order
    if columns is None:
        columns = sorted(vnf_pm_job.keys())
    return tuple([vnf_pm_job[key] for key in columns])



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from tackerclient.osc import utils as tacker_osc_utils


def vnf_package_obj(attrs=None, onboarded_state=False):
    """Create a fake vnf package.

    :param Dictionary attrs:
        A dictionary with all attributes
    :return:
        A FakeVnfPackage dict
    """
    attrs = attrs or {}

    # Set default attributes.
    fake_vnf_package = {"id": "60a6ac16-b50d-4e92-964b-b3cf98c7cf5c",
                        "_links": {"self": {"href": "string"},
                                   "packageContent": {"href": "string"}
                                   },
                        "onboardingState": "CREATED",
                        "operationalState": "DISABLED",
                        "usageState": "NOT_IN_USE",
                        "userDefinedData": {'key': 'value'}}

    if onboarded_state:
        fake_vnf_package = {"id": "60a6ac16-b50d-4e92-964b-b3cf98c7cf5c",
                            "vnfdId": "string",
                            "vnfProvider": "string",
                            "vnfProductName": "string",
                            "vnfSoftwareVersion": "string",
                            "vnfdVersion": "string",
                            "softwareImages": [
                                {
                                    "id": "string",
                                    "name": "string",
                                    "provider": "string",
                                    "version": "string",
                                    "checksum": {
                                        "algorithm": "string",
                                        "hash": "string"
                                    },
                                    "containerFormat": "AKI",
                                    "diskFormat": "AKI",
                                    "createdAt": "2015-06-03T18:49:19.000000",
                                    "minDisk": '0',
                                    "minRam": '0',
                                    "size": '0',
                                    "userMetadata": {},
                                    "imagePath": "string"
                                }
                            ],
                            "checksum": {
                                "algorithm": "string",
                                "hash": "string"
                            },
                            "onboardingState": "ONBOARDED",
                            "operationalState": "ENABLED",
                            "usageState": "IN_USE",
                            "userDefinedData": {'key': 'value'},
                            "_links": {
                                "self": {
                                    "href": "string"
                                },
                                "vnfd": {
                                    "href": "string"
                                },
                                "packageContent": {
                                    "href": "string"
                                }
                            },
                            "additionalArtifacts": [
                                {
                                    "artifactPath": "string",
                                    "metadata": {},
                                    "checksum": {
                                        "algorithm": "string",
                                        "hash": "string"
                                    }
                                }]
                            }

    # Overwrite default attributes.
    fake_vnf_package.update(attrs)
    return fake_vnf_package


def get_vnf_package_data(vnf_package_obj, **kwargs):
    """Get the vnf package data from a FakeVnfPackage dict object.

    :param vnf_package_obj:
        A FakeVnfPackage dict object
    :return:
        A list which may include the following values:
        [{'packageContent': {'href': 'string'}, 'self': {'href': 'string'},
        'vnfd': {'href': 'string'}}, '60a6ac16-b50d-4e92-964b-b3cf98c7cf5c',
        'CREATED', 'DISABLED', 'NOT_IN_USE', {'Test_key': 'Test_value'}]
    """
    complex_attributes = ['softwareImages', 'checksum', '_links',
                          'userDefinedData', 'additionalArtifacts']
    for attribute in complex_attributes:
        if vnf_package_obj.get(attribute):
            vnf_package_obj.update(
                {attribute: tacker_osc_utils.FormatComplexDataColumn(
                    vnf_package_obj[attribute])})

    if kwargs.get('list_action'):
        # In case of List VNF packages we get empty string as data for
        # 'vnfProductName' if onboardingState is CREATED. Hence to match
        # up with actual data we are adding here empty string.
        if not vnf_package_obj.get('vnfProductName'):
            vnf_package_obj['vnfProductName'] = ''

    if kwargs.get('columns'):
        # return the list of data as per column order
        return tuple([vnf_package_obj[key] for key in kwargs.get('columns')])

    return tuple([vnf_package_obj[key] for key in sorted(
        vnf_package_obj.keys())])


def create_vnf_packages(count=2, onboarded_vnf_package=False):
    """Create multiple fake vnf packages.

    :param Dictionary attrs:
        A dictionary with all attributes
    :param int count:
        The number of vnf_packages to fake
    :return:
        A list of fake vnf packages dictionary
    """
    vnf_packages = []
    for i in range(0, count):
        unique_id = uuidutils.generate_uuid()
        vnf_packages.append(vnf_package_obj(
            attrs={'id': unique_id}, onboarded_state=onboarded_vnf_package))
    return {'vnf_packages': vnf_packages}


def get_fake_update_vnf_package_obj(arglist):
    fake_update_vnf_package_dict = {}
    if '--user-data' in arglist:
        fake_update_vnf_package_dict.update(
            {"userDefinedData": {'Test_key': 'Test_value'}})
    if '--operational-state' in arglist:
        fake_update_vnf_package_dict.update({
            "operationalState": "DISABLED",
        })

    return fake_update_vnf_package_dict



# Copyright 2012 OpenStack Foundation.
# Copyright 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import logging
import re
import time

import requests
from urllib import parse as urlparse

from tackerclient import client
from tackerclient.common import exceptions
from tackerclient.common import serializer
from tackerclient.common import utils
from tackerclient.i18n import _

_logger = logging.getLogger(__name__)
DEFAULT_DESC_LENGTH = 25
DEFAULT_ERROR_REASON_LENGTH = 100
STATUS_CODE_MAP = {
    400: "badRequest",
    401: "unauthorized",
    403: "forbidden",
    404: "itemNotFound",
    405: "badMethod",
    409: "conflictingRequest",
    413: "overLimit",
    415: "badMediaType",
    429: "overLimit",
    501: "notImplemented",
    503: "serviceUnavailable"}


def exception_handler_v10(status_code, error_content):
    """Exception handler for API v1.0 client.

    This routine generates the appropriate Tacker exception according to
    the contents of the response body.

    :param status_code: HTTP error status code
    :param error_content: deserialized body of error response
    """
    etsi_error_content = error_content
    error_dict = None
    if isinstance(error_content, dict):
        error_dict = error_content.get('TackerError')
        if not error_dict:
            error_content = error_content.get(STATUS_CODE_MAP.get(status_code),
                                              'tackerFault')
    # Find real error type
    bad_tacker_error_flag = False
    if error_dict:
        # If Tacker key is found, it will definitely contain
        # a 'message' and 'type' keys?
        try:
            error_type = error_dict['type']
            error_message = error_dict['message']
            if error_dict['detail']:
                error_message += "\n" + error_dict['detail']
        except Exception:
            bad_tacker_error_flag = True
        if not bad_tacker_error_flag:
            # If corresponding exception is defined, use it.
            client_exc = getattr(exceptions, '%sClient' % error_type, None)
            # Otherwise look up per status-code client exception
            if not client_exc:
                client_exc = exceptions.HTTP_EXCEPTION_MAP.get(status_code)
            if client_exc:
                raise client_exc(message=error_message,
                                 status_code=status_code)
            else:
                raise exceptions.TackerClientException(
                    status_code=status_code, message=error_message)
        else:
            raise exceptions.TackerClientException(status_code=status_code,
                                                   message=error_dict)
    else:
        message = None
        if isinstance(error_content, dict):
            message = error_content.get('message')
        if message:
            raise exceptions.TackerClientException(status_code=status_code,
                                                   message=message)
        # ETSI error response
        if isinstance(etsi_error_content, dict):
            if etsi_error_content.get('status') and \
                etsi_error_content.get('detail'):
                message = etsi_error_content.get('detail')
                raise exceptions.TackerClientException(status_code=status_code,
                                                       message=message)

    # If we end up here the exception was not a tacker error
    msg = "%s-%s" % (status_code, error_content)
    raise exceptions.TackerClientException(status_code=status_code,
                                           message=msg)


class APIParamsCall(object):
    """A Decorator to support formating and tenant overriding and filters."""

    def __init__(self, function):
        self.function = function

    def __get__(self, instance, owner):
        def with_params(*args, **kwargs):
            _format = instance.format
            if 'format' in kwargs:
                instance.format = kwargs['format']
            ret = self.function(instance, *args, **kwargs)
            instance.format = _format
            return ret
        return with_params


class ClientBase(object):
    """Client for the OpenStack Tacker v1.0 API.

    :param string username: Username for authentication. (optional)
    :param string user_id: User ID for authentication. (optional)
    :param string password: Password for authentication. (optional)
    :param string token: Token for authentication. (optional)
    :param string tenant_name: Tenant name. (optional)
    :param string tenant_id: Tenant id. (optional)
    :param string auth_strategy: 'keystone' by default, 'noauth' for no
                                 authentication against keystone. (optional)
    :param string auth_url: Keystone service endpoint for authorization.
    :param string service_type: Network service type to pull from the
                                keystone catalog (e.g. 'network') (optional)
    :param string endpoint_type: Network service endpoint type to pull from the
                                 keystone catalog (e.g. 'publicURL',
                                 'internalURL', or 'adminURL') (optional)
    :param string region_name: Name of a region to select when choosing an
                               endpoint from the service catalog.
    :param string endpoint_url: A user-supplied endpoint URL for the tacker
                            service.  Lazy-authentication is possible for API
                            service calls if endpoint is set at
                            instantiation.(optional)
    :param integer timeout: Allows customization of the timeout for client
                            http requests. (optional)
    :param bool insecure: SSL certificate validation. (optional)
    :param bool log_credentials: Allow for logging of passwords or not.
                                 Defaults to False. (optional)
    :param string ca_cert: SSL CA bundle file to use. (optional)
    :param integer retries: How many times idempotent (GET, PUT, DELETE)
                            requests to Tacker server should be retried if
                            they fail (default: 0).
    :param bool raise_errors: If True then exceptions caused by connection
                              failure are propagated to the caller.
                              (default: True)
    :param session: Keystone client auth session to use. (optional)
    :param auth: Keystone auth plugin to use. (optional)

    """

    # API has no way to report plurals, so we have to hard code them
    # This variable should be overridden by a child class.
    EXTED_PLURALS = {}

    def __init__(self, **kwargs):
        """Initialize a new client for the Tacker v1.0 API."""
        super(ClientBase, self).__init__()
        self.retries = kwargs.pop('retries', 0)
        self.raise_errors = kwargs.pop('raise_errors', True)
        self.httpclient = client.construct_http_client(**kwargs)
        self.version = '1.0'
        self.format = 'json'
        self.action_prefix = "/v%s" % (self.version)
        self.retry_interval = 1
        self.rel = None
        self.params = None
        self.accept = None

    def _handle_fault_response(self, status_code, response_body):
        # Create exception with HTTP status code and message
        _logger.debug("Error message: %s", response_body)
        # Add deserialized error message to exception arguments
        try:
            des_error_body = self.deserialize(response_body, status_code)
        except Exception:
            # If unable to deserialized body it is probably not a
            # Tacker error
            des_error_body = {'message': response_body}
        # Raise the appropriate exception
        exception_handler_v10(status_code, des_error_body)

    def build_action(self, action):
        action += ".%s" % self.format
        action = self.action_prefix + action
        return action

    def _build_params_query(self, params=None):
        flag_params = []
        keyval_params = {}
        for key, value in params.items():
            if value is None:
                flag_params.append(key)
            else:
                keyval_params[key] = value

        flags_encoded = utils.safe_encode_list(flag_params) \
            if flag_params else ""
        keyval_encoded = utils.safe_encode_dict(keyval_params) \
            if keyval_params else ""

        query = ""
        for flag in flags_encoded:
            query = query + urlparse.quote_plus(flag) + '&'
        query = query + urlparse.urlencode(keyval_encoded, doseq=1)
        return query.strip('&')

    def do_request(self, method, action, body=None, headers=None, params=None):
        action = self.build_action(action)
        # Add format and tenant_id
        if type(params) is dict and params:
            query = self._build_params_query(params)
            action += '?' + query

        if body or body == {}:
            body = self.serialize(body)

        if headers is None:
            # self.httpclient.do_request is not accept 'headers=None'.
            headers = {}

        resp, replybody = self.httpclient.do_request(
            action, method, body=body, headers=headers,
            content_type=self.content_type(), accept=self.accept)

        if 'application/zip' == resp.headers.get('Content-Type'):
            self.format = 'zip'
        elif 'text/plain' == resp.headers.get('Content-Type'):
            self.format = 'text'
        elif 'artifacts' in action:
            self.format = 'any'
        else:
            self.format = 'json'

        url = None
        rel = None

        link = resp.headers.get('Link', None)
        if link is not None:
            url = re.findall('<(.*)>', link)[0]
            rel = re.findall('rel="(.*)"', link)[0]

        if rel == 'next':
            self.rel = 'next'
            query_str = urlparse.urlparse(url).query
            self.params = urlparse.parse_qs(query_str)

        status_code = resp.status_code
        if status_code in (requests.codes.ok,
                           requests.codes.created,
                           requests.codes.accepted,
                           requests.codes.no_content):
            return self.deserialize(replybody, status_code)
        else:
            if not replybody:
                replybody = resp.reason
            self._handle_fault_response(status_code, replybody)

    def get_auth_info(self):
        return self.httpclient.get_auth_info()

    def serialize(self, data):
        """Serializes a dictionary JSON.

        A dictionary with a single key can be passed and it can contain any
        structure.
        """
        if data is None:
            return None
        elif self.format in ('zip', 'text'):
            return data
        elif type(data) is dict:
            return serializer.Serializer().serialize(data, 'application/json')
        else:
            raise Exception(_("Unable to serialize object of type = '%s'") %
                            type(data))

    def deserialize(self, data, status_code):
        """Deserializes an JSON string into a dictionary."""
        if status_code in (204, 202) or self.format in ('zip', 'text', 'any'):
            return data
        return serializer.Serializer().deserialize(
            data, 'application/json')['body']

    def content_type(self, _format=None):
        """Returns the mime-type for either 'json, 'text', or 'zip'.

        Defaults to the currently set format.
        """
        _format = _format or self.format
        if self.format == 'text':
            return "text/plain"
        elif self.format == 'both':
            return "text/plain,application/zip"
        else:
            return "application/%s" % (_format)

    def retry_request(self, method, action, body=None,
                      headers=None, params=None):
        """Call do_request with the default retry configuration.

        Only idempotent requests should retry failed connection attempts.
        :raises ConnectionFailed: if the maximum # of retries is exceeded
        """
        max_attempts = self.retries + 1
        for i in range(max_attempts):
            try:
                return self.do_request(method, action, body=body,
                                       headers=headers, params=params)
            except exceptions.ConnectionFailed:
                # Exception has already been logged by do_request()
                if i < self.retries:
                    _logger.debug('Retrying connection to Tacker service')
                    time.sleep(self.retry_interval)
                elif self.raise_errors:
                    raise

        if self.retries:
            msg = (_("Failed to connect to Tacker server after %d attempts")
                   % max_attempts)
        else:
            msg = _("Failed to connect Tacker server")

        raise exceptions.ConnectionFailed(reason=msg)

    def delete(self, action, body=None, headers=None, params=None):
        return self.retry_request("DELETE", action, body=body,
                                  headers=headers, params=params)

    def get(self, action, body=None, headers=None, params=None):
        return self.retry_request("GET", action, body=body,
                                  headers=headers, params=params)

    def post(self, action, body=None, headers=None, params=None):
        # Do not retry POST requests to avoid the orphan objects problem.
        return self.do_request("POST", action, body=body,
                               headers=headers, params=params)

    def put(self, action, body=None, headers=None, params=None):
        return self.retry_request("PUT", action, body=body,
                                  headers=headers, params=params)

    def patch(self, action, body=None, headers=None, params=None):
        self.format = 'merge-patch+json'
        self.accept = 'json'
        return self.retry_request("PATCH", action, body=body,
                                  headers=headers, params=params)

    def list(self, collection, path, retrieve_all=True, headers=None,
             **params):
        if retrieve_all:
            res = []
            for r in self._pagination(collection, path, headers, **params):
                if type(r) is list:
                    res.extend(r)
                else:
                    res.extend(r[collection])
            return {collection: res} if collection else res
        else:
            return self._pagination(collection, path, headers, **params)

    def _pagination(self, collection, path, headers, **params):
        if params.get('page_reverse', False):
            linkrel = 'previous'
        else:
            linkrel = 'next'
        next = True
        while next:
            self.rel = None
            res = self.get(path, headers=headers, params=params)
            yield res
            next = False
            try:
                if type(res) is list:
                    if self.rel == 'next':
                        params = self.params
                        next = True

                else:
                    for link in res['%s_links' % collection]:
                        if link['rel'] == linkrel:
                            query_str = urlparse.urlparse(link['href']).query
                            params = urlparse.parse_qs(query_str)
                            next = True
                            break
            except KeyError:
                break


class LegacyClient(ClientBase):

    vims_path = '/vims'
    vim_path = '/vims/%s'

    # API has no way to report plurals, so we have to hard code them
    # EXTED_PLURALS = {}

    @APIParamsCall
    def show_vim(self, vim, **_params):
        return self.get(self.vim_path % vim, params=_params)

    _VIM = "vim"

    @APIParamsCall
    def create_vim(self, body):
        return self.post(self.vims_path, body=body)

    @APIParamsCall
    def delete_vim(self, vim):
        return self.delete(self.vim_path % vim)

    @APIParamsCall
    def update_vim(self, vim, body):
        return self.put(self.vim_path % vim, body=body)

    @APIParamsCall
    def list_vims(self, retrieve_all=True, **_params):
        return self.list('vims', self.vims_path, retrieve_all, **_params)


class VnfPackageClient(ClientBase):
    """Client for vnfpackage APIs.

    Purpose of this class is to create required request url for vnfpackage
    APIs.
    """

    vnfpackages_path = '/vnfpkgm/v1/vnf_packages'
    vnfpackage_path = '/vnfpkgm/v1/vnf_packages/%s'
    vnfpackage_vnfd_path = '/vnfpkgm/v1/vnf_packages/%s/vnfd'
    vnfpackage_download_path = '/vnfpkgm/v1/vnf_packages/%s/package_content'
    vnfpakcage_artifact_path = '/vnfpkgm/v1/vnf_packages/%(id)s/artifacts/' \
                               '%(artifact_path)s'

    def build_action(self, action):
        return action

    @APIParamsCall
    def create_vnf_package(self, body):
        return self.post(self.vnfpackages_path, body=body)

    @APIParamsCall
    def list_vnf_packages(self, retrieve_all=True, **_params):
        vnf_packages = self.list("vnf_packages", self.vnfpackages_path,
                                 retrieve_all, **_params)
        return vnf_packages

    @APIParamsCall
    def show_vnf_package(self, vnf_package, **_params):
        return self.get(self.vnfpackage_path % vnf_package, params=_params)

    @APIParamsCall
    def delete_vnf_package(self, vnf_package):
        return self.delete(self.vnfpackage_path % vnf_package)

    @APIParamsCall
    def upload_vnf_package(self, vnf_package, file_data=None, **attrs):
        if attrs.get('url'):
            json = {'addressInformation': attrs.get('url')}
            for key in ['userName', 'password']:
                if attrs.get(key):
                    json.update({key: attrs.get(key)})
            return self.post(
                '{base_path}/{id}/package_content/upload_from_uri'.format(
                    id=vnf_package, base_path=self.vnfpackages_path),
                body=json)
        else:
            self.format = 'zip'
            self.accept = 'json'
            return self.put('{base_path}/{id}/package_content'.format(
                id=vnf_package,
                base_path=self.vnfpackages_path),
                body=file_data)

    @APIParamsCall
    def download_vnf_package(self, vnf_package):
        self.format = 'zip'
        return self.get(self.vnfpackage_download_path % vnf_package)

    @APIParamsCall
    def download_vnfd_from_vnf_package(self, vnf_package, accept):
        """Read VNFD of an on-boarded VNF Package.

        :param vnf_package: The value can be either the ID of a vnf package
                            or a :class:`~openstack.nfv_orchestration.v1.
                            vnf_package` instance.
        :param accept: Valid values are 'text/plain', 'application/zip' and
                       'both'. According to these values 'Accept' header will
                        be set as 'text/plain', 'application/zip',
                       'text/plain,application/zip' respectively.

        :returns: If the VNFD is implemented in the form of multiple files,
                  a ZIP file embedding these files shall be returned.
                  If the VNFD is implemented as a single file, either that
                  file or a ZIP file embedding that file shall be returned.
        """
        if accept == 'text/plain':
            self.format = 'text'
        elif accept == 'application/zip':
            self.format = 'zip'
        else:
            self.format = 'both'
        return self.get(self.vnfpackage_vnfd_path % vnf_package)

    @APIParamsCall
    def download_artifact_from_vnf_package(self, vnf_package, artifact_path):
        return self.get(self.vnfpakcage_artifact_path %
                        {'id': vnf_package, 'artifact_path': artifact_path})

    @APIParamsCall
    def update_vnf_package(self, vnf_package, body):
        return self.patch(self.vnfpackage_path % vnf_package, body=body)


class VnfLCMClient(ClientBase):
    """Client for vnflcm APIs.

    Purpose of this class is to create required request url for vnflcm
    APIs.
    """

    def __init__(self, api_version, **kwargs):
        super(VnfLCMClient, self).__init__(**kwargs)
        self.headers = {'Version': '1.3.0'}
        sol_api_version = 'v1'
        if api_version == '2':
            self.headers = {'Version': '2.0.0'}
            sol_api_version = 'v2'

        self.vnf_instances_path = (
            '/vnflcm/{}/vnf_instances'.format(sol_api_version))
        self.vnf_instance_path = (
            '/vnflcm/{}/vnf_instances/%s'.format(sol_api_version))
        self.vnf_lcm_op_occurrences_path = (
            '/vnflcm/{}/vnf_lcm_op_occs'.format(sol_api_version))
        self.vnf_lcm_op_occs_path = (
            '/vnflcm/{}/vnf_lcm_op_occs/%s'.format(sol_api_version))
        self.lccn_subscriptions_path = (
            '/vnflcm/{}/subscriptions'.format(sol_api_version))
        self.lccn_subscription_path = (
            '/vnflcm/{}/subscriptions/%s'.format(sol_api_version))

    def build_action(self, action):
        return action

    @APIParamsCall
    def create_vnf_instance(self, body):
        return self.post(self.vnf_instances_path, body=body,
                         headers=self.headers)

    @APIParamsCall
    def show_vnf_instance(self, vnf_id, **_params):
        return self.get(self.vnf_instance_path % vnf_id,
                        headers=self.headers, params=_params)

    @APIParamsCall
    def list_vnf_instances(self, retrieve_all=True, **_params):
        vnf_instances = self.list(None, self.vnf_instances_path,
                                  retrieve_all, headers=self.headers,
                                  **_params)
        return vnf_instances

    @APIParamsCall
    def instantiate_vnf_instance(self, vnf_id, body):
        return self.post((self.vnf_instance_path + "/instantiate") % vnf_id,
                         body=body, headers=self.headers)

    @APIParamsCall
    def heal_vnf_instance(self, vnf_id, body):
        return self.post((self.vnf_instance_path + "/heal") % vnf_id,
                         body=body, headers=self.headers)

    @APIParamsCall
    def terminate_vnf_instance(self, vnf_id, body):
        return self.post((self.vnf_instance_path + "/terminate") % vnf_id,
                         body=body, headers=self.headers)

    @APIParamsCall
    def delete_vnf_instance(self, vnf_id):
        return self.delete(self.vnf_instance_path % vnf_id,
                           headers=self.headers)

    @APIParamsCall
    def update_vnf_instance(self, vnf_id, body):
        return self.patch(self.vnf_instance_path % vnf_id, body=body,
                          headers=self.headers)

    @APIParamsCall
    def scale_vnf_instance(self, vnf_id, body):
        return self.post((self.vnf_instance_path + "/scale") % vnf_id,
                         body=body, headers=self.headers)

    @APIParamsCall
    def rollback_vnf_instance(self, occ_id):
        return self.post((self.vnf_lcm_op_occs_path + "/rollback") % occ_id,
                         headers=self.headers)

    @APIParamsCall
    def cancel_vnf_instance(self, occ_id, body):
        return self.post((self.vnf_lcm_op_occs_path + "/cancel") % occ_id,
                         body=body)

    @APIParamsCall
    def fail_vnf_instance(self, occ_id):
        return self.post((self.vnf_lcm_op_occs_path + "/fail") % occ_id,
                         headers=self.headers)

    @APIParamsCall
    def change_ext_conn_vnf_instance(self, vnf_id, body):
        return self.post((self.vnf_instance_path + "/change_ext_conn") %
                         vnf_id, body=body, headers=self.headers)

    @APIParamsCall
    def change_vnfpkg_vnf_instance(self, vnf_id, body):
        # NOTE: it is only supported by V2-API.
        if self.vnf_instance_path.split('/')[2] == 'v2':
            return self.post((self.vnf_instance_path + "/change_vnfpkg") %
                             vnf_id, body=body, headers=self.headers)
        else:
            raise exceptions.UnsupportedCommandVersion(version='1')

    @APIParamsCall
    def retry_vnf_instance(self, occ_id):
        return self.post((self.vnf_lcm_op_occs_path + "/retry") % occ_id,
                         headers=self.headers)

    @APIParamsCall
    def list_vnf_lcm_op_occs(self, retrieve_all=True, **_params):
        vnf_lcm_op_occs = self.list(None, self.vnf_lcm_op_occurrences_path,
                                    retrieve_all, headers=self.headers,
                                    **_params)
        return vnf_lcm_op_occs

    @APIParamsCall
    def show_vnf_lcm_op_occs(self, occ_id):
        return self.get(self.vnf_lcm_op_occs_path % occ_id,
                        headers=self.headers)

    @APIParamsCall
    def create_lccn_subscription(self, body):
        return self.post(self.lccn_subscriptions_path, body=body,
                         headers=self.headers)

    @APIParamsCall
    def delete_lccn_subscription(self, subsc_id):
        return self.delete(self.lccn_subscription_path % subsc_id,
                           headers=self.headers)

    @APIParamsCall
    def list_lccn_subscriptions(self, retrieve_all=True, **_params):
        subscriptions = self.list(None, self.lccn_subscriptions_path,
                                  retrieve_all, headers=self.headers,
                                  **_params)
        return subscriptions

    @APIParamsCall
    def show_lccn_subscription(self, subsc_id):
        return self.get(self.lccn_subscription_path % subsc_id,
                        headers=self.headers)

    @APIParamsCall
    def show_vnf_lcm_versions(self, major_version):
        if major_version is None:
            path = "/vnflcm/api_versions"
        else:
            path = "/vnflcm/{}/api_versions".format(major_version)
        # NOTE: This may be called with any combination of
        # --os-tacker-api-verson:[1, 2] and major_version:[None, 1, 2].
        # Specifying "headers={'Version': '2.0.0'}" is most simple to
        # make all cases OK.
        return self.get(path, headers={'Version': '2.0.0'})


class VnfFMClient(ClientBase):
    headers = {'Version': '1.3.0'}
    vnf_fm_alarms_path = '/vnffm/v1/alarms'
    vnf_fm_alarm_path = '/vnffm/v1/alarms/%s'
    vnf_fm_subs_path = '/vnffm/v1/subscriptions'
    vnf_fm_sub_path = '/vnffm/v1/subscriptions/%s'

    def build_action(self, action):
        return action

    @APIParamsCall
    def list_vnf_fm_alarms(self, retrieve_all=True, **_params):
        vnf_fm_alarms = self.list(
            "vnf_fm_alarms", self.vnf_fm_alarms_path, retrieve_all,
            headers=self.headers, **_params)
        return vnf_fm_alarms

    @APIParamsCall
    def show_vnf_fm_alarm(self, vnf_fm_alarm_id):
        return self.get(
            self.vnf_fm_alarm_path % vnf_fm_alarm_id, headers=self.headers)

    @APIParamsCall
    def update_vnf_fm_alarm(self, vnf_fm_alarm_id, body):
        return self.patch(
            self.vnf_fm_alarm_path % vnf_fm_alarm_id, body=body,
            headers=self.headers)

    @APIParamsCall
    def create_vnf_fm_sub(self, body):
        return self.post(
            self.vnf_fm_subs_path, body=body, headers=self.headers)

    @APIParamsCall
    def list_vnf_fm_subs(self, retrieve_all=True, **_params):
        vnf_fm_subs = self.list("vnf_fm_subs", self.vnf_fm_subs_path,
                                retrieve_all, headers=self.headers, **_params)
        return vnf_fm_subs

    @APIParamsCall
    def show_vnf_fm_sub(self, vnf_fm_sub_id):
        return self.get(
            self.vnf_fm_sub_path % vnf_fm_sub_id, headers=self.headers)

    @APIParamsCall
    def delete_vnf_fm_sub(self, vnf_fm_sub_id):
        return self.delete(
            self.vnf_fm_sub_path % vnf_fm_sub_id, headers=self.headers)


class VnfPMClient(ClientBase):
    headers = {'Version': '2.1.0'}
    vnf_pm_jobs_path = '/vnfpm/v2/pm_jobs'
    vnf_pm_job_path = '/vnfpm/v2/pm_jobs/%s'
    vnf_pm_reports_path = '/vnfpm/v2/pm_jobs/%(job_id)s/reports/%(report_id)s'
    vnf_pm_thresholds_path = '/vnfpm/v2/thresholds'
    vnf_pm_threshold_path = '/vnfpm/v2/thresholds/%s'

    def build_action(self, action):
        return action

    @APIParamsCall
    def create_vnf_pm_job(self, body):
        return self.post(
            self.vnf_pm_jobs_path, body=body, headers=self.headers)

    @APIParamsCall
    def list_vnf_pm_jobs(self, retrieve_all=True, **_params):
        vnf_pm_jobs = self.list(
            "vnf_pm_jobs", self.vnf_pm_jobs_path, retrieve_all,
            headers=self.headers, **_params)
        return vnf_pm_jobs

    @APIParamsCall
    def show_vnf_pm_job(self, vnf_pm_job_id):
        return self.get(
            self.vnf_pm_job_path % vnf_pm_job_id, headers=self.headers)

    @APIParamsCall
    def update_vnf_pm_job(self, vnf_pm_job_id, body):
        return self.patch(
            self.vnf_pm_job_path % vnf_pm_job_id, body=body,
            headers=self.headers)

    @APIParamsCall
    def delete_vnf_pm_job(self, vnf_pm_job_id):
        return self.delete(
            self.vnf_pm_job_path % vnf_pm_job_id, headers=self.headers)

    @APIParamsCall
    def show_vnf_pm_report(self, vnf_pm_job_id, vnf_pm_report_id):
        return self.get(
            self.vnf_pm_reports_path % {
                'job_id': vnf_pm_job_id, 'report_id': vnf_pm_report_id
            }, headers=self.headers)

    @APIParamsCall
    def create_vnf_pm_threshold(self, body):
        return self.post(
            self.vnf_pm_thresholds_path, body=body, headers=self.headers)

    @APIParamsCall
    def list_vnf_pm_thresholds(self, retrieve_all=True, **_params):
        return self.list(
            "vnf_pm_thresholds", self.vnf_pm_thresholds_path, retrieve_all,
            headers=self.headers, **_params)

    @APIParamsCall
    def show_vnf_pm_threshold(self, vnf_pm_threshold_id):
        return self.get(
            self.vnf_pm_threshold_path % vnf_pm_threshold_id,
            headers=self.headers)

    @APIParamsCall
    def update_vnf_pm_threshold(self, vnf_pm_threshold_id, body):
        return self.patch(
            self.vnf_pm_threshold_path % vnf_pm_threshold_id, body=body,
            headers=self.headers)

    @APIParamsCall
    def delete_vnf_pm_threshold(self, vnf_pm_threshold_id):
        return self.delete(
            self.vnf_pm_threshold_path % vnf_pm_threshold_id,
            headers=self.headers)


class Client(object):
    """Unified interface to interact with multiple apps of tacker service.

    This class is a single entry point to interact with legacy tacker apis and
    vnf packages apis.

        Example::

        from tackerclient.v1_0 import client
        tacker = client.Client(username=USER,
                                password=PASS,
                                tenant_name=TENANT_NAME,
                                auth_url=KEYSTONE_URL)

        vnf_package = tacker.create_vnf_package(...)
        nsd = tacker.create_nsd(...)

    """

    def __init__(self, **kwargs):
        api_version = kwargs.pop('api_version', '1')
        self.vnf_lcm_client = VnfLCMClient(api_version, **kwargs)
        self.vnf_fm_client = VnfFMClient(**kwargs)
        self.vnf_pm_client = VnfPMClient(**kwargs)
        self.vnf_package_client = VnfPackageClient(**kwargs)
        self.legacy_client = LegacyClient(**kwargs)

    # LegacyClient methods

    def delete(self, action, body=None, headers=None, params=None):
        return self.legacy_client.delete(action, body=body, headers=headers,
                                         params=params)

    def get(self, action, body=None, headers=None, params=None):
        return self.legacy_client.get(action, body=body, headers=headers,
                                      params=params)

    def post(self, action, body=None, headers=None, params=None):
        return self.legacy_client.post(action, body=body, headers=headers,
                                       params=params)

    def put(self, action, body=None, headers=None, params=None):
        return self.legacy_client.put(action, body=body, headers=headers,
                                      params=params)

    def list(self, collection, path, retrieve_all=True, **params):
        return self.legacy_client.list(collection, path,
                                       retrieve_all=retrieve_all, **params)

    def show_vim(self, vim, **_params):
        return self.legacy_client.show_vim(vim, **_params)

    def create_vim(self, body):
        return self.legacy_client.create_vim(body)

    def delete_vim(self, vim):
        return self.legacy_client.delete_vim(vim)

    def update_vim(self, vim, body):
        return self.legacy_client.update_vim(vim, body)

    def list_vims(self, retrieve_all=True, **_params):
        return self.legacy_client.list_vims(retrieve_all=retrieve_all,
                                            **_params)

    # VnfPackageClient methods

    def create_vnf_package(self, body):
        return self.vnf_package_client.create_vnf_package(body)

    def list_vnf_packages(self, retrieve_all=True, query_parameter=None,
                          **_params):
        return self.vnf_package_client.list_vnf_packages(
            retrieve_all=retrieve_all, **_params)

    def show_vnf_package(self, vnf_package, **_params):
        return self.vnf_package_client.show_vnf_package(vnf_package, **_params)

    def upload_vnf_package(self, vnf_package, file_data=None, **_params):
        return self.vnf_package_client.upload_vnf_package(
            vnf_package, file_data=file_data, **_params)

    def delete_vnf_package(self, vnf_package):
        return self.vnf_package_client.delete_vnf_package(vnf_package)

    # VnfLCMClient methods.

    def create_vnf_instance(self, body):
        return self.vnf_lcm_client.create_vnf_instance(body)

    def show_vnf_instance(self, vnf_instance, **_params):
        return self.vnf_lcm_client.show_vnf_instance(vnf_instance,
                                                     **_params)

    def list_vnf_instances(self, retrieve_all=True, **_params):
        return self.vnf_lcm_client.list_vnf_instances(
            retrieve_all=retrieve_all, **_params)

    def instantiate_vnf_instance(self, vnf_id, body):
        return self.vnf_lcm_client.instantiate_vnf_instance(vnf_id, body)

    def heal_vnf_instance(self, vnf_id, body):
        return self.vnf_lcm_client.heal_vnf_instance(vnf_id, body)

    def terminate_vnf_instance(self, vnf_id, body):
        return self.vnf_lcm_client.terminate_vnf_instance(vnf_id, body)

    def scale_vnf_instance(self, vnf_id, body):
        return self.vnf_lcm_client.scale_vnf_instance(vnf_id, body)

    def change_ext_conn_vnf_instance(self, vnf_id, body):
        return self.vnf_lcm_client.change_ext_conn_vnf_instance(vnf_id, body)

    def change_vnfpkg_vnf_instance(self, vnf_id, body):
        return self.vnf_lcm_client.change_vnfpkg_vnf_instance(vnf_id, body)

    def delete_vnf_instance(self, vnf_id):
        return self.vnf_lcm_client.delete_vnf_instance(vnf_id)

    def update_vnf_instance(self, vnf_id, body):
        return self.vnf_lcm_client.update_vnf_instance(vnf_id, body)

    def rollback_vnf_instance(self, occ_id):
        return self.vnf_lcm_client.rollback_vnf_instance(occ_id)

    def cancel_vnf_instance(self, occ_id, body):
        return self.vnf_lcm_client.cancel_vnf_instance(occ_id, body)

    def fail_vnf_instance(self, occ_id):
        return self.vnf_lcm_client.fail_vnf_instance(occ_id)

    def retry_vnf_instance(self, occ_id):
        return self.vnf_lcm_client.retry_vnf_instance(occ_id)

    def update_vnf_package(self, vnf_package, body):
        return self.vnf_package_client.update_vnf_package(vnf_package, body)

    def download_vnfd_from_vnf_package(self, vnf_package, accept):
        return self.vnf_package_client.download_vnfd_from_vnf_package(
            vnf_package, accept)

    def download_artifact_from_vnf_package(self, vnf_package, artifact_path):
        return self.vnf_package_client.download_artifact_from_vnf_package(
            vnf_package, artifact_path
        )

    def download_vnf_package(self, vnf_package):
        return self.vnf_package_client.download_vnf_package(vnf_package)

    def list_vnf_lcm_op_occs(self, retrieve_all=True, **_params):
        return self.vnf_lcm_client.list_vnf_lcm_op_occs(
            retrieve_all=retrieve_all, **_params)

    def show_vnf_lcm_op_occs(self, occ_id):
        return self.vnf_lcm_client.show_vnf_lcm_op_occs(occ_id)

    def create_lccn_subscription(self, body):
        return self.vnf_lcm_client.create_lccn_subscription(body)

    def delete_lccn_subscription(self, subsc_id):
        return self.vnf_lcm_client.delete_lccn_subscription(subsc_id)

    def list_lccn_subscriptions(self, retrieve_all=True, **_params):
        return self.vnf_lcm_client.list_lccn_subscriptions(
            retrieve_all=retrieve_all, **_params)

    def show_lccn_subscription(self, subsc_id):
        return self.vnf_lcm_client.show_lccn_subscription(subsc_id)

    def show_vnf_lcm_versions(self, major_version):
        return self.vnf_lcm_client.show_vnf_lcm_versions(major_version)

    # VnfFMClient methods.

    def list_vnf_fm_alarms(self, retrieve_all=True, **_params):
        return self.vnf_fm_client.list_vnf_fm_alarms(
            retrieve_all=retrieve_all, **_params)

    def show_vnf_fm_alarm(self, vnf_fm_alarm_id):
        return self.vnf_fm_client.show_vnf_fm_alarm(vnf_fm_alarm_id)

    def update_vnf_fm_alarm(self, vnf_fm_alarm_id, body):
        return self.vnf_fm_client.update_vnf_fm_alarm(vnf_fm_alarm_id, body)

    def create_vnf_fm_sub(self, body):
        return self.vnf_fm_client.create_vnf_fm_sub(body)

    def list_vnf_fm_subs(self, retrieve_all=True, **_params):
        return self.vnf_fm_client.list_vnf_fm_subs(
            retrieve_all=retrieve_all, **_params)

    def show_vnf_fm_sub(self, vnf_fm_sub_id):
        return self.vnf_fm_client.show_vnf_fm_sub(vnf_fm_sub_id)

    def delete_vnf_fm_sub(self, vnf_fm_sub_id):
        return self.vnf_fm_client.delete_vnf_fm_sub(vnf_fm_sub_id)

    # VnfPMClient methods.

    def create_vnf_pm_job(self, body):
        return self.vnf_pm_client.create_vnf_pm_job(body)

    def list_vnf_pm_jobs(self, retrieve_all=True, **_params):
        return self.vnf_pm_client.list_vnf_pm_jobs(
            retrieve_all=retrieve_all, **_params)

    def show_vnf_pm_job(self, vnf_pm_job_id):
        return self.vnf_pm_client.show_vnf_pm_job(vnf_pm_job_id)

    def update_vnf_pm_job(self, vnf_pm_job_id, body):
        return self.vnf_pm_client.update_vnf_pm_job(vnf_pm_job_id, body)

    def delete_vnf_pm_job(self, vnf_pm_job_id):
        return self.vnf_pm_client.delete_vnf_pm_job(vnf_pm_job_id)

    def show_vnf_pm_report(self, vnf_pm_job_id, vnf_pm_report_id):
        return self.vnf_pm_client.show_vnf_pm_report(
            vnf_pm_job_id, vnf_pm_report_id)

    def create_vnf_pm_threshold(self, body):
        return self.vnf_pm_client.create_vnf_pm_threshold(body)

    def list_vnf_pm_thresholds(self, retrieve_all=True, **_params):
        return self.vnf_pm_client.list_vnf_pm_thresholds(
            retrieve_all=retrieve_all, **_params)

    def show_vnf_pm_threshold(self, vnf_pm_threshold_id):
        return self.vnf_pm_client.show_vnf_pm_threshold(vnf_pm_threshold_id)

    def update_vnf_pm_threshold(self, vnf_pm_threshold_id, body):
        return self.vnf_pm_client.update_vnf_pm_threshold(
            vnf_pm_threshold_id, body)

    def delete_vnf_pm_threshold(self, vnf_pm_threshold_id):
        return self.vnf_pm_client.delete_vnf_pm_threshold(vnf_pm_threshold_id)



# Copyright (C) 2022 Fujitsu
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import os
import sys

from io import StringIO
from oslo_utils.fixture import uuidsentinel
from unittest import mock

from tackerclient.common import exceptions
from tackerclient.osc import utils as tacker_osc_utils
from tackerclient.osc.v2.vnffm import vnffm_sub
from tackerclient.tests.unit.osc import base
from tackerclient.tests.unit.osc.v1.fixture_data import client
from tackerclient.tests.unit.osc.v2 import vnffm_sub_fakes


class TestVnfFmSub(base.FixturedTestCase):
    client_fixture_class = client.ClientFixture

    def setUp(self):
        super(TestVnfFmSub, self).setUp()
        self.url = client.TACKER_URL
        self.header = {'content-type': 'application/json'}
        self.app = mock.Mock()
        self.app_args = mock.Mock()
        self.client_manager = self.cs
        self.app.client_manager.tackerclient = self.client_manager


def _get_columns_vnffm_sub(action=None):
    columns = ['ID', 'Callback Uri']

    if action == 'show' or action == 'create':
        columns.extend(['Filter', 'Links'])

    return columns


@ddt.ddt
class TestCreateVnfFmSub(TestVnfFmSub):

    def setUp(self):
        super(TestCreateVnfFmSub, self).setUp()
        self.create_vnf_fm_sub = vnffm_sub.CreateVnfFmSub(
            self.app, self.app_args, cmd_name='vnffm sub create')

    def test_create_no_args(self):
        self.assertRaises(base.ParserException, self.check_parser,
                          self.create_vnf_fm_sub, [], [])

    @ddt.unpack
    def test_take_action(self):

        param_file = ("./tackerclient/osc/v2/vnffm/samples/"
                      "create_vnf_fm_subscription_param_sample.json")

        arg_list = [param_file]
        verify_list = [('request_file', param_file)]

        parsed_args = self.check_parser(self.create_vnf_fm_sub, arg_list,
                                        verify_list)

        json = vnffm_sub_fakes.vnf_fm_sub_response()
        self.requests_mock.register_uri(
            'POST', os.path.join(self.url, 'vnffm/v1/subscriptions'),
            json=json, headers=self.header)

        actual_columns, data = (
            self.create_vnf_fm_sub.take_action(parsed_args))

        _, attributes = vnffm_sub._get_columns(json)

        self.assertCountEqual(_get_columns_vnffm_sub("create"),
                              actual_columns)
        self.assertListItemsEqual(vnffm_sub_fakes.get_vnffm_sub_data(
            json, columns=attributes), data)


class TestListVnfFmSub(TestVnfFmSub):

    def setUp(self):
        super(TestListVnfFmSub, self).setUp()
        self.list_vnffm_sub = vnffm_sub.ListVnfFmSub(
            self.app, self.app_args, cmd_name='vnffm sub list')

    def test_take_action(self):
        vnffm_subs_obj = vnffm_sub_fakes.create_vnf_fm_subs(
            count=3)
        parsed_args = self.check_parser(self.list_vnffm_sub, [], [])
        self.requests_mock.register_uri(
            'GET', os.path.join(self.url, 'vnffm/v1/subscriptions'),
            json=vnffm_subs_obj, headers=self.header)

        actual_columns, data = self.list_vnffm_sub.take_action(parsed_args)

        _, columns = tacker_osc_utils.get_column_definitions(
            vnffm_sub._ATTR_MAP, long_listing=True)

        expected_data = []
        for vnffm_sub_obj_idx in vnffm_subs_obj:
            expected_data.append(vnffm_sub_fakes.get_vnffm_sub_data(
                vnffm_sub_obj_idx, columns=columns))

        self.assertCountEqual(_get_columns_vnffm_sub(action='list'),
                              actual_columns)
        self.assertCountEqual(expected_data, list(data))

    def test_take_action_with_filter(self):
        vnffm_subs_obj = vnffm_sub_fakes.create_vnf_fm_subs(
            count=3)
        parsed_args = self.check_parser(
            self.list_vnffm_sub,
            ["--filter", '(eq,callbackUri,/nfvo/notify/alarm)'],
            [('filter', '(eq,callbackUri,/nfvo/notify/alarm)')])
        self.requests_mock.register_uri(
            'GET', os.path.join(
                self.url,
                'vnffm/v1/subscriptions?'
                'filter=(eq,callbackUri,/nfvo/notify/alarm)'),
            json=vnffm_subs_obj, headers=self.header)

        actual_columns, data = self.list_vnffm_sub.take_action(parsed_args)

        _, columns = tacker_osc_utils.get_column_definitions(
            vnffm_sub._ATTR_MAP, long_listing=True)

        expected_data = []
        for vnffm_sub_obj_idx in vnffm_subs_obj:
            expected_data.append(vnffm_sub_fakes.get_vnffm_sub_data(
                vnffm_sub_obj_idx, columns=columns))

        self.assertCountEqual(_get_columns_vnffm_sub(action='list'),
                              actual_columns)
        self.assertListItemsEqual(expected_data, list(data))

    def test_take_action_with_incorrect_filter(self):

        parsed_args = self.check_parser(
            self.list_vnffm_sub,
            ["--filter", '(callbackUri)'],
            [('filter', '(callbackUri)')])

        url = os.path.join(
            self.url,
            'vnffm/v1/subscriptions?filter=(callbackUri)')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=400, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.list_vnffm_sub.take_action,
                          parsed_args)

    def test_take_action_internal_server_error(self):

        parsed_args = self.check_parser(
            self.list_vnffm_sub,
            ["--filter", '(eq,callbackUri,/nfvo/notify/alarm)'],
            [('filter', '(eq,callbackUri,/nfvo/notify/alarm)')])

        url = os.path.join(
            self.url,
            'vnffm/v1/subscriptions?'
            'filter=(eq,callbackUri,/nfvo/notify/alarm)')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=500, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.list_vnffm_sub.take_action,
                          parsed_args)


class TestShowVnfFmSub(TestVnfFmSub):

    def setUp(self):
        super(TestShowVnfFmSub, self).setUp()
        self.show_vnf_fm_subs = vnffm_sub.ShowVnfFmSub(
            self.app, self.app_args, cmd_name='vnffm sub show')

    def test_take_action(self):
        """Test of take_action()"""
        vnffm_sub_obj = vnffm_sub_fakes.vnf_fm_sub_response()

        arg_list = [vnffm_sub_obj['id']]
        verify_list = [('vnf_fm_sub_id', vnffm_sub_obj['id'])]

        # command param
        parsed_args = self.check_parser(
            self.show_vnf_fm_subs, arg_list, verify_list)
        url = os.path.join(
            self.url,
            'vnffm/v1/subscriptions',
            vnffm_sub_obj['id'])

        self.requests_mock.register_uri(
            'GET', url, headers=self.header, json=vnffm_sub_obj)

        columns, _ = (self.show_vnf_fm_subs.take_action(parsed_args))

        self.assertCountEqual(_get_columns_vnffm_sub('show'),
                              columns)

    def test_take_action_vnf_fm_sub_id_not_found(self):
        """Test if vnf-lcm-op-occ-id does not find."""
        arg_list = [uuidsentinel.vnf_fm_sub_id]
        verify_list = [('vnf_fm_sub_id', uuidsentinel.vnf_fm_sub_id)]

        # command param
        parsed_args = self.check_parser(
            self.show_vnf_fm_subs, arg_list, verify_list)

        url = os.path.join(
            self.url,
            'vnffm/v1/subscriptions',
            uuidsentinel.vnf_fm_sub_id)
        self.requests_mock.register_uri(
            'GET', url, headers=self.header, status_code=404, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.show_vnf_fm_subs.take_action,
                          parsed_args)

    def test_take_action_internal_server_error(self):
        """Test for internal server error."""
        arg_list = [uuidsentinel.vnf_fm_sub_id]
        verify_list = [('vnf_fm_sub_id', uuidsentinel.vnf_fm_sub_id)]

        # command param
        parsed_args = self.check_parser(
            self.show_vnf_fm_subs, arg_list, verify_list)

        url = os.path.join(
            self.url,
            'vnffm/v1/subscriptions',
            uuidsentinel.vnf_fm_sub_id)
        self.requests_mock.register_uri(
            'GET', url, headers=self.header, status_code=500, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.show_vnf_fm_subs.take_action,
                          parsed_args)


class TestDeleteVnfFmSub(TestVnfFmSub):

    def setUp(self):
        super(TestDeleteVnfFmSub, self).setUp()
        self.delete_vnf_fm_sub = vnffm_sub.DeleteVnfFmSub(
            self.app, self.app_args, cmd_name='vnffm sub delete')

        # Vnf Fm subscription to delete
        self.vnf_fm_subs = vnffm_sub_fakes.create_vnf_fm_subs(count=3)

    def _mock_request_url_for_delete(self, index):
        url = os.path.join(self.url, 'vnffm/v1/subscriptions',
                           self.vnf_fm_subs[index]['id'])

        self.requests_mock.register_uri('DELETE', url,
                                        headers=self.header, json={})

    def test_delete_one_vnf_fm_sub(self):
        arg_list = [self.vnf_fm_subs[0]['id']]
        verify_list = [('vnf_fm_sub_id',
                       [self.vnf_fm_subs[0]['id']])]

        parsed_args = self.check_parser(self.delete_vnf_fm_sub, arg_list,
                                        verify_list)

        self._mock_request_url_for_delete(0)
        sys.stdout = buffer = StringIO()
        result = self.delete_vnf_fm_sub.take_action(parsed_args)
        self.assertIsNone(result)
        self.assertEqual(
            (f"VNF FM subscription '{self.vnf_fm_subs[0]['id']}' "
             f"deleted successfully"), buffer.getvalue().strip())

    def test_delete_multiple_vnf_fm_sub(self):
        arg_list = []
        for obj in self.vnf_fm_subs:
            arg_list.append(obj['id'])
        verify_list = [('vnf_fm_sub_id', arg_list)]
        parsed_args = self.check_parser(self.delete_vnf_fm_sub, arg_list,
                                        verify_list)
        for i in range(0, 3):
            self._mock_request_url_for_delete(i)
        sys.stdout = buffer = StringIO()
        result = self.delete_vnf_fm_sub.take_action(parsed_args)
        self.assertIsNone(result)
        self.assertEqual('All specified VNF FM subscriptions are deleted '
                         'successfully', buffer.getvalue().strip())

    def test_delete_multiple_vnf_fm_sub_exception(self):
        arg_list = [
            self.vnf_fm_subs[0]['id'],
            'xxxx-yyyy-zzzz',
            self.vnf_fm_subs[1]['id'],
        ]
        verify_list = [('vnf_fm_sub_id', arg_list)]
        parsed_args = self.check_parser(self.delete_vnf_fm_sub,
                                        arg_list, verify_list)

        self._mock_request_url_for_delete(0)

        url = os.path.join(self.url, 'vnffm/v1/subscriptions',
                           'xxxx-yyyy-zzzz')
        self.requests_mock.register_uri(
            'GET', url, exc=exceptions.ConnectionFailed)

        self._mock_request_url_for_delete(1)
        exception = self.assertRaises(exceptions.CommandError,
                                      self.delete_vnf_fm_sub.take_action,
                                      parsed_args)

        self.assertEqual('Failed to delete 1 of 3 VNF FM subscriptions.',
                         exception.message)



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from io import StringIO
import os
import sys
from unittest import mock

from tackerclient import client as root_client
from tackerclient.common import exceptions
from tackerclient.osc.v1.vnflcm import vnflcm
from tackerclient.tests.unit.osc import base
from tackerclient.tests.unit.osc.v1.fixture_data import client
from tackerclient.tests.unit.osc.v1 import test_vnflcm
from tackerclient.tests.unit.osc.v1 import vnflcm_fakes
from tackerclient.tests.unit.test_cli10 import MyResp
from tackerclient.v1_0 import client as proxy_client


class TestVnfLcmV2(base.FixturedTestCase):
    client_fixture_class = client.ClientFixture
    api_version = '2'

    def setUp(self):
        super(TestVnfLcmV2, self).setUp()

    def test_client_v2(self):
        self.assertEqual(self.cs.vnf_lcm_client.headers,
                         {'Version': '2.0.0'})
        self.assertEqual(self.cs.vnf_lcm_client.vnf_instances_path,
                         '/vnflcm/v2/vnf_instances')
        # check of other paths is omitted.


class TestUpdateVnfLcmV2(test_vnflcm.TestVnfLcm):
    api_version = '2'

    def setUp(self):
        super(TestUpdateVnfLcmV2, self).setUp()
        self.update_vnf_lcm = vnflcm.UpdateVnfLcm(
            self.app, self.app_args, cmd_name='vnflcm modify')

    def test_take_action_check_content_type(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = ('./tackerclient/osc/v1/vnflcm/samples/'
                             'update_vnf_instance_param_sample.json')

        arglist = [vnf_instance['id'], '--I', sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('I', sample_param_file)]

        # command param
        parsed_args = self.check_parser(
            self.update_vnf_lcm, arglist, verifylist)
        url = os.path.join(self.url, 'vnflcm/v2/vnf_instances',
                           vnf_instance['id'])

        self.requests_mock.register_uri(
            'PATCH', url, headers=self.header, json={})

        sys.stdout = buffer = StringIO()

        with mock.patch.object(root_client.HTTPClient,
                               'do_request') as mock_req:
            headers = {'Content-Type': 'application/json'}
            mock_req.return_value = (MyResp(202, headers=headers), None)
            self.update_vnf_lcm.take_action(parsed_args)
            # check content_type
            mock_req.assert_called_once_with(
                f'/vnflcm/v2/vnf_instances/{vnf_instance["id"]}', 'PATCH',
                body=mock.ANY, headers=mock.ANY,
                content_type='application/merge-patch+json', accept='json')

        actual_message = buffer.getvalue().strip()
        expected_message = f'Update vnf:{vnf_instance["id"]}'
        self.assertEqual(expected_message, actual_message)


class TestChangeVnfPkgVnfLcm(test_vnflcm.TestVnfLcm):
    api_version = '2'

    def setUp(self):
        super(TestChangeVnfPkgVnfLcm, self).setUp()
        self.change_vnfpkg_vnf_lcm = vnflcm.ChangeVnfPkgVnfLcm(
            self.app, self.app_args,
            cmd_name='vnflcm change-vnfpkg')

    def test_take_action(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = ("./tackerclient/osc/v2/vnflcm/samples/"
                             "change_vnfpkg_vnf_instance_param_sample.json")

        arglist = [vnf_instance['id'], sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('request_file', sample_param_file)]

        # command param
        parsed_args = self.check_parser(self.change_vnfpkg_vnf_lcm,
                                        arglist,
                                        verifylist)

        url = os.path.join(self.url, 'vnflcm/v2/vnf_instances',
                           vnf_instance['id'], 'change_vnfpkg')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, json={})

        sys.stdout = buffer = StringIO()
        with mock.patch.object(proxy_client.ClientBase,
                               '_handle_fault_response') as m:
            self.change_vnfpkg_vnf_lcm.take_action(parsed_args)
            # check no fault response is received
            self.assertNotCalled(m)
            self.assertEqual(
                ('Change Current VNF Package for VNF Instance {0} '
                 'has been accepted.'.format(vnf_instance['id'])),
                buffer.getvalue().strip())

    def test_take_action_vnf_instance_not_found(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = ("./tackerclient/osc/v1/vnflcm/samples/"
                             "change_vnfpkg_vnf_instance_param_sample.json")
        arglist = [vnf_instance['id'], sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('request_file', sample_param_file)]

        # command param
        parsed_args = self.check_parser(self.change_vnfpkg_vnf_lcm,
                                        arglist,
                                        verifylist)

        url = os.path.join(self.url, 'vnflcm/v2/vnf_instances',
                           vnf_instance['id'], 'change_vnfpkg')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=404, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.change_vnfpkg_vnf_lcm.take_action,
                          parsed_args)

    def test_take_action_param_file_not_exists(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = "./not_exists.json"
        arglist = [vnf_instance['id'], sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('request_file', sample_param_file)]

        # command param
        parsed_args = self.check_parser(
            self.change_vnfpkg_vnf_lcm,
            arglist,
            verifylist)

        ex = self.assertRaises(
            exceptions.InvalidInput,
            self.change_vnfpkg_vnf_lcm.take_action,
            parsed_args)

        expected_msg = ("Invalid input: "
                        "User does not have read privileges to it")
        self.assertEqual(expected_msg, str(ex))

    @mock.patch("os.open")
    @mock.patch("os.access")
    def test_take_action_invalid_format_param_file(self, mock_access,
                                                   mock_open):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = "./invalid_param_file.json"
        arglist = [vnf_instance['id'], sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('request_file', sample_param_file)]

        mock_open.return_value = "invalid_json_data"
        mock_access.return_value = True
        # command param
        parsed_args = self.check_parser(self.change_vnfpkg_vnf_lcm,
                                        arglist,
                                        verifylist)

        ex = self.assertRaises(
            exceptions.InvalidInput,
            self.change_vnfpkg_vnf_lcm.take_action,
            parsed_args)

        expected_msg = "Failed to load parameter file."
        self.assertIn(expected_msg, str(ex))



# Copyright (C) 2013 Yahoo! Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
import fixtures
import io
import logging
import os
import re
import sys
import testtools
from testtools import matchers
from unittest import mock

from keystoneclient import session

from tackerclient.common import clientmanager
from tackerclient import shell as openstack_shell


DEFAULT_USERNAME = 'username'
DEFAULT_PASSWORD = 'password'
DEFAULT_TENANT_ID = 'tenant_id'
DEFAULT_TENANT_NAME = 'tenant_name'
DEFAULT_AUTH_URL = 'http://127.0.0.1:5000/v1.0/'
DEFAULT_TOKEN = '3bcc3d3a03f44e3d8377f9247b0ad155'
DEFAULT_URL = 'http://tacker.example.org:9890/'
DEFAULT_API_VERSION = '1.0'


class ShellTest(testtools.TestCase):

    FAKE_ENV = {
        'OS_USERNAME': DEFAULT_USERNAME,
        'OS_PASSWORD': DEFAULT_PASSWORD,
        'OS_TENANT_ID': DEFAULT_TENANT_ID,
        'OS_TENANT_NAME': DEFAULT_TENANT_NAME,
        'OS_AUTH_URL': DEFAULT_AUTH_URL}

    # Patch os.environ to avoid required auth info.
    def setUp(self):
        super(ShellTest, self).setUp()
        for var in self.FAKE_ENV:
            self.useFixture(
                fixtures.EnvironmentVariable(
                    var, self.FAKE_ENV[var]))

    def shell(self, argstr, check=False):
        orig = (sys.stdout, sys.stderr)
        clean_env = {}
        _old_env, os.environ = os.environ, clean_env.copy()
        try:
            sys.stdout = io.StringIO()
            sys.stderr = io.StringIO()
            _shell = openstack_shell.TackerShell(DEFAULT_API_VERSION)
            _shell.run(argstr.split())
        except SystemExit:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.assertEqual(exc_value.code, 0)
        finally:
            stdout = sys.stdout.getvalue()
            stderr = sys.stderr.getvalue()
            sys.stdout.close()
            sys.stderr.close()
            sys.stdout, sys.stderr = orig
            os.environ = _old_env
        return stdout, stderr

    def test_run_unknown_command(self):
        self.useFixture(fixtures.FakeLogger(level=logging.DEBUG))
        stdout, stderr = self.shell('fake', check=True)
        self.assertFalse(stdout)
        self.assertEqual("Unknown command ['fake']", stderr.strip())

    def test_help(self):
        required = 'usage:'
        help_text, stderr = self.shell('help')
        self.assertThat(
            help_text,
            matchers.MatchesRegex(required))
        self.assertFalse(stderr)

    def test_help_on_subcommand(self):
        required = [
            '.*?^usage: .* vim-list']
        stdout, stderr = self.shell('help vim-list')
        for r in required:
            self.assertThat(
                stdout,
                matchers.MatchesRegex(r, re.DOTALL | re.MULTILINE))
        self.assertFalse(stderr)

    def test_help_command(self):
        required = 'usage:'
        help_text, stderr = self.shell('help vim-create')
        self.assertThat(
            help_text,
            matchers.MatchesRegex(required))
        self.assertFalse(stderr)

    def test_unknown_auth_strategy(self):
        self.useFixture(fixtures.FakeLogger(level=logging.DEBUG))
        stdout, stderr = self.shell('--os-auth-strategy fake '
                                    'vim-list')
        self.assertFalse(stdout)

    def test_auth(self):
        with mock.patch.object(openstack_shell.TackerShell,
                               'run_subcommand'), \
                mock.patch.object(session, 'Session'), \
                mock.patch.object(clientmanager, 'ClientManager') as mock_cmgr:

            shell = openstack_shell.TackerShell(DEFAULT_API_VERSION)
            shell.options = mock.Mock()
            auth_session = shell._get_keystone_session()

            cmdline = ('--os-username test '
                       '--os-password test '
                       '--os-tenant-name test '
                       '--os-auth-url http://127.0.0.1:5000/ '
                       '--os-auth-strategy keystone vim-list')
            shell.authenticate_user()
            shell.run(cmdline.split())

        mock_cmgr.assert_called_with(
            raise_errors=False, retries=0, timeout=None,
            token='', url='', auth_url='http://127.0.0.1:5000/',
            tenant_name='test', tenant_id='tenant_id',
            username='test', user_id='',
            password='test', region_name='',
            api_version={'nfv-orchestration': '1.0'},
            auth_strategy='keystone',
            service_type='nfv-orchestration',
            endpoint_type='publicURL', insecure=False, ca_cert=None,
            log_credentials=True, session=auth_session, auth=auth_session.auth)

    def test_build_option_parser(self):
        tacker_shell = openstack_shell.TackerShell(DEFAULT_API_VERSION)
        result = tacker_shell.build_option_parser('descr', DEFAULT_API_VERSION)
        self.assertIsInstance(result, argparse.ArgumentParser)

    @mock.patch.object(openstack_shell.TackerShell, 'run')
    def test_main_with_unicode(self, mock_run):
        mock_run.return_value = 0
        unicode_text = '\u7f51\u7edc'
        argv = ['net-list', unicode_text, unicode_text.encode('utf-8')]
        ret = openstack_shell.main(argv=argv)
        mock_run.assert_called_once_with(['net-list', unicode_text,
                                          unicode_text])
        self.assertEqual(0, ret)

    def test_endpoint_option(self):
        shell = openstack_shell.TackerShell(DEFAULT_API_VERSION)
        parser = shell.build_option_parser('descr', DEFAULT_API_VERSION)

        # Neither $OS_ENDPOINT_TYPE nor --endpoint-type
        namespace = parser.parse_args([])
        self.assertEqual('publicURL', namespace.endpoint_type)

        # --endpoint-type but not $OS_ENDPOINT_TYPE
        namespace = parser.parse_args(['--endpoint-type=admin'])
        self.assertEqual('admin', namespace.endpoint_type)

    def test_endpoint_environment_variable(self):
        fixture = fixtures.EnvironmentVariable("OS_ENDPOINT_TYPE",
                                               "public")
        self.useFixture(fixture)

        shell = openstack_shell.TackerShell(DEFAULT_API_VERSION)
        parser = shell.build_option_parser('descr', DEFAULT_API_VERSION)

        # $OS_ENDPOINT_TYPE but not --endpoint-type
        namespace = parser.parse_args([])
        self.assertEqual("public", namespace.endpoint_type)

        # --endpoint-type and $OS_ENDPOINT_TYPE
        namespace = parser.parse_args(['--endpoint-type=admin'])
        self.assertEqual('admin', namespace.endpoint_type)
